'''
PlanarDiagram
=============

Classes for working with planar diagram notation of knot diagrams.

See individual class documentation for more details.

API documentation
~~~~~~~~~~~~~~~~~
'''

from __future__ import print_function

import numpy as n


class PlanarDiagram(list):
    '''A class for containing and manipulating planar diagrams.

    Just provides convenient display and conversion methods for now.
    In the future, will support simplification.

    Shorthand input may be of the form ``X_1,4,2,5 X_3,6,4,1 X_5,2,6,3``.
    This is (should be?) the same as returned by repr.

    Parameters
    ----------
    crossings : array-like or string or GaussCode
        The list of crossings in the diagram, which will be converted
        to an internal planar diagram representation. Currently these are
        mostly converted via a GaussCode instance, so in addition to the
        shorthand any array-like supported by
        :class:`~pyknotid.representations.gausscode.GaussCode` may be used.
    '''

    def __init__(self, crossings=''):
        from pyknotid.representations import gausscode
        if isinstance(crossings, str):
            self.extend(shorthand_to_crossings(crossings))
        elif isinstance(crossings, gausscode.GaussCode):
            self.extend(gausscode_to_crossings(crossings))
        else:
            self.extend(gausscode_to_crossings(
                gausscode.GaussCode(crossings)))

    def __str__(self):
        lenstr = 'PD with {0}: '.format(len(self))
        return lenstr + ' '.join([str(crossing) for crossing in self])

    def __repr__(self):
        return self.__str__()

    def as_mathematica(self):
        '''
        Returns a mathematica code representation of self, usable in the
        mathematica knot tools.
        '''
        s = 'PD['
        s = s + ', '.join(crossing.as_mathematica() for crossing in self)
        return s + ']'

    def as_spherogram(self):
        '''
        Get a planar diagram class from the Spherogram module, which
        can be used to access SnapPy's manifold tools.

        This method requires that spherogram and SnapPy are installed.
        '''
        from snappy import Link
        from spherogram import Crossing, Link
        scs = [Crossing() for crossing in self]

        indices = {}
        for i in range(len(self)):
            c = self[i]
            for j in range(len(c)):
                number = c[j]
                if number in indices:
                    otheri, otherj = indices.pop(number)
                    scs[i][j] = scs[otheri][otherj]
                else:
                    indices[number] = (i, j)
        return Link(scs)

    def as_networkx_extended(self):
        '''(internal use only) Returns a networkx Graph along with extra
        information about the crossings.
        '''
        # print('pd is', self)
        import networkx as nx
        edges = []
        cache = {}
        heights = {}
        edge_directions = []
        from collections import defaultdict
        intermediate_edges_by_node = defaultdict(lambda : [None, None, None, None])
        intermediate_node_index = len(self)
        for node_index, crossing in enumerate(self):
            for crossing_index, arc_number in enumerate(crossing):
                if arc_number in cache:
                    other_node_index, other_height, other_crossing_index = cache[arc_number]
                    # edges.append([other_node_index, node_index])
                    edges.append([other_node_index, intermediate_node_index])
                    edges.append([node_index, intermediate_node_index])

                    height = index_height(crossing_index)

                    if crossing.is_outgoing(arc_number):
                        # heights[node_index, other_node_index, arc_number] = (height, other_height)
                        heights[node_index, intermediate_node_index, arc_number] = (height, 0.)
                        heights[intermediate_node_index, other_node_index, arc_number] = (0., other_height)

                        # edge_directions.append((node_index, other_node_index, arc_number))
                        edge_directions.append((node_index, intermediate_node_index, arc_number))
                        edge_directions.append((intermediate_node_index, other_node_index, arc_number))
                    else:
                        # heights[other_node_index, node_index, arc_number] = (other_height, height)
                        heights[other_node_index, intermediate_node_index, arc_number] = (other_height, 0.)
                        heights[intermediate_node_index, node_index, arc_number] = (0., height)

                        # edge_directions.append((other_node_index, node_index, arc_number))
                        edge_directions.append((other_node_index, intermediate_node_index, arc_number))
                        edge_directions.append((intermediate_node_index, node_index, arc_number))

                    intermediate_edges_by_node[node_index][crossing_index] = intermediate_node_index
                    intermediate_edges_by_node[other_node_index][other_crossing_index] = intermediate_node_index

                    intermediate_node_index += 1 
                    
                else:
                    cache[arc_number] = (node_index, index_height(crossing_index), crossing_index)

        for node, intermediate_nodes in intermediate_edges_by_node.items():
            for i, intermediate in enumerate(intermediate_nodes):
                next_intermediate = intermediate_nodes[(i+1) % 4]
                edges.append([intermediate, next_intermediate])

        # print('intermediates are', intermediate_edges_by_node)
        g = nx.Graph()
        g.add_nodes_from(list(range(len(self)))*3)
        g.add_edges_from(edges)

        # print('possible heights:')
        # for key, value in sorted(heights.items()):
        #     print(key, value)


        seen = set()
        duplicates = list()
        for edge in edges:
            edge = tuple(edge)
            if edge in seen:
                duplicates.append(sorted(edge))
            seen.add(edge)

        return g, duplicates, heights, sorted(edge_directions)[0]

    def as_networkx(self):
        '''Get a networkx graph representing the planar diagram, where
        each node is a crossing and each edge is an arc. This is a
        non-directed non-multi graph; where two arcs join the same crossing,
        they are represented as a single edge, but information about
        duplicates is returned alongside the graph.

        Returns
        -------
        g : Graph
            The networkx graph
        duplicates : list
            A list of tuples representing nodes joined by multiple edges.
        heights : dict
            A dictionary of (start, end, arc_number) graph edges,
            containing the start and end height of each edge.
        first_edge : tuple
            The first edge in the graph, including (start, end, arc_number).
        '''
        # print('pd is', self)
        import networkx as nx
        edges = []
        cache = {}
        heights = {}
        edge_directions = []
        for node_index, crossing in enumerate(self):
            for crossing_index, arc_number in enumerate(crossing):
                if arc_number in cache:
                    other_node_index, other_height = cache[arc_number]
                    edges.append([other_node_index, node_index])

                    height = index_height(crossing_index)

                    if crossing.is_outgoing(arc_number):
                        # print('node {} {}, crossing {} is outgoing to {}'.format(
                        #     node_index, crossing, arc_number, other_node_index))
                        heights[node_index, other_node_index, arc_number] = (height, other_height)
                        edge_directions.append((node_index, other_node_index, arc_number))
                    else:
                        # print('node {} {}, crossing {} is incoming from {}'.format(
                            # node_index, crossing, arc_number, other_node_index))
                        heights[other_node_index, node_index, arc_number] = (other_height, height)
                        edge_directions.append((other_node_index, node_index, arc_number))
                    
                else:
                    cache[arc_number] = node_index, index_height(crossing_index)

        g = nx.Graph()
        g.add_nodes_from(range(len(self)))
        g.add_edges_from(edges)

        # print('possible heights:')
        # for key, value in sorted(heights.items()):
        #     print(key, value)


        seen = set()
        duplicates = list()
        for edge in edges:
            edge = tuple(edge)
            if edge in seen:
                duplicates.append(sorted(edge))
            seen.add(edge)

        return g, duplicates, heights, sorted(edge_directions)[0]
            
def index_height(index):
    '''Returns the height based on the index of the crossing in an entry
    of a planar diagram; the 0th and 2nd indices are under crossings,
    and the 1st and 3rd are over crossings.

    '''
    if index in (0, 2):
        return -1.
    return 1.

class Crossing(list):
    '''
    A single crossing in a planar diagram. Each :class:`PlanarDiagram`
    is a list of these.

    Parameters
    ----------
    a : int or None
        The first entry in the list of lines meeting at this Crossing.
    b : int or None
        The second entry in the list of lines meeting at this Crossing.
    c : int or None
        The third entry in the list of lines meeting at this Crossing.
    d : int or None
        The fourth entry in the list of lines meeting at this Crossing.
    '''

    def __init__(self, a=None, b=None, c=None, d=None):
        super(Crossing, self).__init__()
        self.extend([a, b, c, d])

    def valid(self):
        '''
        True if all intersecting lines are not None.
        '''
        if all([entry is not None for entry in self]):
            return True
        return False

    def components(self):
        '''
        Returns a de-duplicated list of lines intersecting at this Crossing.

        :rtype: list
        '''
        return list(set(self))

    def __str__(self):
        return 'X_{{{0},{1},{2},{3}}}'.format(
            self[0], self[1], self[2], self[3])

    def __repr__(self):
        return self.__str__()

    def as_mathematica(self):
        '''
        Get a string of mathematica code that can represent the Crossing
        in mathematica's knot library.

        The mathematica code won't be valid if any lines of self are None.

        :rtype: str
        '''
        return 'X[{}, {}, {}, {}]'.format(
            self[0], self[1], self[2], self[3])

    def __hash__(self):
        return tuple(self).__hash__()

    def update_line_number(self, old, new):
        '''
        Replaces all instances of the given line number in self.

        Parameters
        ----------
        old : int
            The old line number
        new : int
            The number to replace it with
        '''
        for i in range(4):
            if self[i] == old:
                self[i] = new

    def is_incoming(self, index):
        if index not in self:
            raise ValueError('arc index doesn\'t intersect crossing')
        other_index = self[(self.index(index) + 2) % 4]
        if other_index == 1 and index != 2:  # => index is the final arc
            return True
        if index == 1 and other_index != 2:  # => other_index is the final arc
            return False
        if other_index > index:
            return True
        return False

    def is_outgoing(self, index):
        return not self.is_incoming(index)

def shorthand_to_crossings(s):
    '''
    Takes a planar diagram shorthand string, and returns a list of
    :class:`Crossing`s.
    '''
    crossings = []
    cs = s.split(' ')
    for entry in cs:
        entry = entry.split('_')
        if entry[0] == 'X':
            a, b, c, d = [int(j) for j in entry[1].split(',')]
            crossings.append(Crossing(a, b, c, d))
        elif entry[0] == 'P':
            a, b = [int(j) for j in entry[1].split(',')]
            crossings.append(Point(a, b))
    return crossings


def gausscode_to_crossings(gc):
    cl = gc._gauss_code
    crossings = []
    incomplete_crossings = {}
    line_lengths = [len(line) for line in cl]
    total_lines = sum(line_lengths)
    line_indices = [1] + list(n.cumsum(line_lengths)[:-1] + 1)

    curline = 1
    for i, line in enumerate(cl):
        curline = line_indices[i]
        for index, over, clockwise in line:
            if index in incomplete_crossings:
                crossing = incomplete_crossings.pop(index)
            else:
                crossing = Crossing()

            inline = curline
            curline += 1
            if curline >= (line_indices[i] + line_lengths[i]):
                curline = line_indices[i]
            outline = curline

            if over == -1:
                crossing[0] = inline
                crossing[2] = outline
                crossings.append(crossing)
            else:
                if clockwise == 1:
                    crossing[3] = inline
                    crossing[1] = outline
                else:
                    crossing[1] = inline
                    crossing[3] = outline

            if not crossing.valid():
                incomplete_crossings[index] = crossing

    return crossings



#!/usr/bin/env python2

from __future__ import print_function, division

import argparse
import sys
import numpy as np

from os import path
sys.path.append(path.split(path.dirname(path.abspath(__file__)))[0])

from pyknotid.spacecurves.knot import Knot


def knot_from_file(filen):
    with open(filen, 'r') as fileh:
        text = fileh.read()

    return knot_from_text(text)
    

def knot_from_text(text):

    points = []
    for line in text.split('\n'):
        values = line.split(' ')
        if len(values) == 1 and len(values[0]) == 0:
            continue
        if len(values) != 3:
            raise ValueError('Invalid text passed.')
        points.append([float(v) for v in values])

    k = Knot(points, verbose=False)
    return k


def main():
    parser = argparse.ArgumentParser(
        description='Analyse the knot type of a curve in a data file')

    parser.add_argument('filename', help='The file to load')

    parser.add_argument('--identify', dest='identify', action='store_true',
                        help=('Look up the knot in the pyknotid database, '
                              'using its invariants. This may take a few '
                              'seconds on the first run, and can only return '
                              'knots with the same invariants, not necessarily '
                              'perfect matches.'))
    parser.add_argument('--no-identify', dest='identify', action='store_false')
    parser.set_defaults(identify=True)

    parser.add_argument('--rotate', dest='rotate', action='store_true',
                        help=('Rotate the space curve randomly before '
                              'identifying. Usually a good idea, and '
                              'defaults to True.'))
    parser.add_argument('--no-rotate', dest='rotate', action='store_false')
    parser.set_defaults(rotate=True)

    parser.add_argument('--writhe', action='store_true',
                        help=('If passed, will also print an approximation '
                              'for the 3D writhe and average crossing number '
                              'of the knot.'))
    parser.add_argument('--projections', type=int, default=100,
                        help=('The number of projections to average over '
                              'when calculating writhe.'))

    parser.add_argument('--with-vassiliev-3', action='store_true',
                        help=('If passed, also calculates the Vassiliev '
                              'invariant of third order (per Polyak and '
                              'Viro'))

    args = parser.parse_args(sys.argv[1:])

    k = knot_from_file(args.filename)

    print('Loaded line with {} points'.format(len(k.points)))
    
    if args.rotate:
        k.rotate()
    else:
        k.rotate((0.001, 0.0023, 0.0052))

    invariants = ['determinant',
                  'vassiliev_degree_2',
                  'alexander_at_roots']

    print('  determinant:', k.determinant())
    print('  Vassiliev degree 2:', k.vassiliev_degree_2())
    if args.with_vassiliev_3:
        print('  Vassiliev degree 3:', k.vassiliev_degree_3())
    print('  Alexander at roots 2,3,4:', k.alexander_at_root((2, 3, 4)))

    if args.identify:
        ## Identification taken from knot.py (and extended)
        identify_kwargs = {}
        for root in (2, 3, 4):
            identify_kwargs[
                'alex_imag_{}'.format(root)] = k.alexander_at_root(root)

        import sympy as sym
        poly = k.alexander_polynomial(sym.var('t'))
        identify_kwargs['alexander'] = poly

        if len(k.gauss_code()) < 16:
            identify_kwargs['max_crossings'] = len(k.gauss_code())

        if args.with_vassiliev_3:
            identify_kwargs['vassiliev_3'] = k.vassiliev_degree_3()

        from pyknotid.catalogue.identify import from_invariants
        identities = from_invariants(**identify_kwargs)
        if len(identities) < 10:
            print('  Possible identities:', identities)
        else:
            print('  First 10 of {} possible identities:'.format(len(identities)),
                identities[:10])

    import warnings
    warnings.filterwarnings('ignore')
    if args.writhe:
        acn, writhe = k._writhe_and_crossing_numbers(args.projections)
        print('  average crossing number:', acn)
        print('  writhe:', writhe)

if __name__ == "__main__":
    main()
    



from setuptools import setup, find_packages
from distutils.extension import Extension

from os.path import join, sep, dirname
from os import walk, environ
import glob
import re


package_data = {}

def recursively_include(results, directory, patterns):
    for root, subfolders, files in walk(directory):
        for fn in files:
            if not any([glob.fnmatch.fnmatch(fn, pattern) for pattern in patterns]):
                continue
            filename = join(root, fn)
            directory = 'pyknotid'
            if directory not in results:
                results[directory] = []
            results[directory].append(join(*filename.split(sep)[1:]))

recursively_include(package_data, 'pyknotid',
                    ['*.tmpl', '*.pov', '*.pyx', '*.pxd',
                     '*.py',
                     ])

# Build cython components if possible
try:
    from Cython.Build import cythonize
    import numpy
except ImportError:
    print('Cython or numpy could not be imported, so cythonised calculation '
          'functions will not be built. pyknotid will use Python-only '
          'routines instead. These are slower, but will return the '
          'same result.')
    print('To build the cython components, install cython and numpy and rebuild '
          'pyknotid.')
    ext_modules = []
    include_dirs = []
else:
    ext_modules = [
            Extension("pyknotid.spacecurves.chelpers", ["pyknotid/spacecurves/chelpers.pyx"],
                    libraries=["m"]),
            Extension("pyknotid.spacecurves.ccomplexity", ["pyknotid/spacecurves/ccomplexity.pyx"],
                    libraries=["m"]),
            Extension("pyknotid.simplify.coctree", ["pyknotid/simplify/coctree.pyx"],
                    libraries=["m"]),
            Extension("pyknotid.cinvariants", ["pyknotid/cinvariants.pyx"],
                    libraries=["m"]),
            ]
    ext_modules = cythonize(ext_modules)
    include_dirs = [numpy.get_include()]

pyknotid_init_filen = join(dirname(__file__), 'pyknotid', '__init__.py')
version = None
try:
    with open(pyknotid_init_filen) as fileh:
        lines = fileh.readlines()
except IOError:
    pass
else:
    for line in lines:
        line = line.strip()
        if line.startswith('__version__ = '):
            matches = re.findall(r'["\'].+["\']', line)
            if matches:
                version = matches[0].strip("'").strip('"')
                break
if version is None:
    raise Exception('Error: version could not be loaded from {}'.format(pyknotid_init_filen))

if 'READTHEDOCS' in environ and environ['READTHEDOCS'] == 'True':
    print('Installing for doc only')
    install_requires=['numpy', 'peewee', 'vispy', 'sympy']
else:
    install_requires=['numpy', 'networkx', 'planarity',
                      'peewee', 'vispy', 'sympy', 'appdirs',
                      'requests', 'tqdm'],

long_description = '''
Pyknotid
========

Python (and optional Cython) modules for detecting and measuring
knotting and linking. pyknotid can analyse space-curves, i.e. sets of
points in three-dimensions, or can parse standard topological
representations of knot diagrams.

A graphical interface to some of these tools is available online at
`Knot ID <http://inclem.net/knotidentifier>`__.

pyknotid was developed as part of the Leverhulme Trust Research
Programme Grant RP2013-K-009: Scientific Properties of Complex Knots
(SPOCK), a collaboration between the University of Bristol and Durham
University in the UK. For more information, see the `SPOCK homepage
<http://www.maths.dur.ac.uk/spock/index.html/>`__.

If you use pyknotid in your research, please `cite us
<http://pyknotid.readthedocs.io/en/latest/sources/about.html#cite-us>`__.

Questions or comments are welcome, please email alexander.taylor@bristol.ac.uk.

Documentation
-------------

pyknotid is documented online at `readthedocs
<http://pyknotid.readthedocs.io/en/latest/sources/overview.html>`__.
'''

setup(
    name='pyknotid',
    version=version,
    description=('Tools for identifying and analysing knots, in space-curves '
                 'or standard topological representations'),
    long_description=long_description,
    author='Alexander Taylor',
    author_email='alexander.taylor@bristol.ac.uk',
    install_requires=install_requires,
    ext_modules=ext_modules,
    include_dirs=include_dirs,
    packages=find_packages(),
    package_data=package_data,
    entry_points={
        'console_scripts': [
            'analyse-knot-file = pyknotid.cli.analyse_knot_file:main',
            'plot-knot = pyknotid.cli.plot_knot:main']
        }
)



'''Virtual Knots
=============

Class for representing a virtual knot and calling useful methods. This
really subclasses GaussCode, not a space curve.

'''


from pyknot2.representations.gausscode import GaussCode


class




import numpy as n
import itertools

def plain_lemniscate(bulbs, strands, repeats, numpoints=200):
    """Return sets of numpoints describing the space curve(s) obtained
    via the lemniscate construction.

    """

    ps = []

    xs = n.linspace(0, 2*n.pi * repeats, numpoints)
    ys = n.linspace(0, 2*n.pi * bulbs * repeats, numpoints)
    angles = n.linspace(0, 2*n.pi*strands, numpoints)

    for i in range(numpoints):
        r = 80 + 50*n.cos(xs[i])
        angle = angles[i]

        realx = r*n.cos(angle)
        realy = r*n.sin(angle)
        realz = 20*n.sin(ys[i])
        ps.append([realx, realy, realz])

    return n.array(ps)
    
def complex_lemniscate(strands=5, repeats=1, numpoints=200):
    '''Return sets of numpoints describing the space curve(s) obtained by
    a multidimensional lemniscate construction.

    '''

    ts = n.linspace(0, 2*n.pi*repeats, numpoints)
    xs = n.sin(2*ts)
    ys = n.sin(3*ts)

    angles = n.linspace(0, 2*n.pi*strands, numpoints)
    
    ps = []
    for i in range(numpoints):
        r = 80 + 50*xs[i]
        angle = angles[i]

        realx = r*n.cos(angle)
        realy = r*n.sin(angle)
        realz = 20*n.sin(ys[i])
        ps.append([realx, realy, realz])

    ps = n.array(ps)
    ps[-1] *= 0.99

    return ps



'''
Planar diagrams
===============

Classes for working with planar diagram notation of knot diagrams.

See individual class documentation for more details.
'''


import numpy as n


class PlanarDiagram(list):
    '''A class for containing and manipulating planar diagrams.

    Just provides convenient display and conversion methods for now.
    In the future, will support simplification.

    Shorthand input may be of the form ``X_1,4,2,5 X_3,6,4,1 X_5,2,6,3``.
    This is (should be?) the same as returned by repr.

    Parameters
    ----------
    crossings : array-like or string or GaussCode
        The list of crossings in the diagram, which will be converted
        to an internal planar diagram representation. Currently these are
        mostly converted via a GaussCode instance, so in addition to the
        shorthand any array-like supported by
        :class:`~pyknot2.representations.gausscode.GaussCode` may be used.
    '''

    def __init__(self, crossings=''):
        from pyknot2.representations import gausscode
        if isinstance(crossings, str):
            self.extend(shorthand_to_crossings(crossings))
        elif isinstance(crossings, gausscode.GaussCode):
            self.extend(gausscode_to_crossings(crossings))
        elif isinstance(crossings, list) and (isinstance(crossings[0], Crossing) or isinstance(crossings[0], Point)):
            self.extend(crossings)
        elif isinstance(crossings, Crossing) or isinstance(crossings, Point):
            self.extend([crossings])
        else:
            self.extend(gausscode_to_crossings(
                gausscode.GaussCode(crossings)))


    def __str__(self):
        number_of_crossings = sum(isinstance(i, Crossing) for i in self)
        number_of_points = sum(isinstance(i, Point) for i in self)
        lenstr = 'PD with {0} crossings and {1} points: \n'.format(
            number_of_crossings, number_of_points)
        return lenstr + ' '.join([str(crossing) for crossing in self])

    def __repr__(self):
        return self.__str__()

    def as_mathematica(self):
        '''
        Returns a mathematica code representation of self, usable in the
        mathematica knot tools.
        '''
        s = 'PD['
        s = s + ', '.join(crossing.as_mathematica() for crossing in self)
        return s + ']'

    def as_spherogram(self):
        '''
        Get a planar diagram class from the Spherogram module, which
        can be used to access SnapPy's manifold tools.

        This method requires that spherogram and SnapPy are installed.
        '''
        from spherogram import Crossing, Link
        scs = [Crossing() for crossing in self]

        indices = {}
        for i in range(len(self)):
            c = self[i]
            for j in range(len(c)):
                number = c[j]
                if number in indices:
                    otheri, otherj = indices.pop(number)
                    scs[i][j] = scs[otheri][otherj]
                else:
                    indices[number] = (i, j)
        return Link(scs)

    def contract_points(self):
        '''
        For appropriately contracting :class: `Points` in a
        :class: `PlanarDiagram` According to the following rules:

        P_a,b P_b,c -> P_a,c
        P_a,b P_a,b -> P_a,a
        '''
        import copy

        pd_crossings = [x for x in self if type(x) == Crossing]
        pd_points = [x for x in self if type(x) == Point]

        pd_trimmed = PlanarDiagram()
        for crossing in pd_crossings:
            pd_trimmed.append(crossing)

        substituted_points = []
        for i in range(len(pd_points)):
            if pd_points[i] not in substituted_points:
                pd_trimmed.append(pd_points[i])
                totally_simplified = False
                while not totally_simplified:
                    pd_trimmed_old = copy.copy(pd_trimmed)
                    for j in range(i+1, len(pd_points)):
                        if (len(pd_trimmed[-1].components()) > 1 and
                                len(pd_points[j].components()) > 1 and
                                pd_points[j] not in substituted_points):
                            # This skips P_a,a situations
                            if pd_trimmed[-1].components() != pd_points[j].components():
                                # This conditions skips (P_a,b P_a,b) situations
                                substituted = False
                                p = 0
                                while substituted is not True and p != 2:
                                    q = 0
                                    while substituted is not True and q != 2:
                                        if pd_trimmed[-1].components()[p] == pd_points[j].components()[q]:
                                            pd_trimmed.append(Point(pd_trimmed[-1].components()[(p+1) % 2],
                                                                    pd_points[j].components()[(q+1) % 2]))
                                            del pd_trimmed[-2]
                                            substituted = True
                                            substituted_points.append(pd_points[j])
                                        q += 1
                                    p += 1
                            else:
                                # This deals with (P_a,b P_a,b) situations
                                pd_trimmed.append(Point(pd_trimmed[-1].components()[0],
                                                        pd_trimmed[-1].components()[0]))
                                del pd_trimmed[-2]
                                substituted_points.append(pd_points[j])
                    if pd_trimmed_old == pd_trimmed:
                        totally_simplified = True
        return pd_trimmed

    def writhe(self):
        crossings = [x for x in self if type(x) == Crossing]
        total_writhe = 0
        for crossing in crossings:
            total_writhe += crossing.sign()
        return total_writhe

    def jones_polynomial(self, **kwargs):
        root_of_unity = kwargs.get('root_of_unity')
        if root_of_unity is not None:
            A = root_of_unity
            A **= 1/4
        else:
            import sympy as sym
            A = sym.var('A')
            q = sym.var('q')
        split_diagram = kauffman_all_crossings(self)
        replacement = -A**2 - A**-2
        jones_precursor = 0
        for diagram in split_diagram:
            jones_precursor += A**diagram[0] * replacement**len(diagram[1])
        jones_poly = (-A**3)**self.writhe() * jones_precursor / replacement
        if root_of_unity is None:
            return sym.expand(sym.simplify(jones_poly)).subs(A, q**0.25)
        else:
            return jones_poly

    def jones_optimised(self, **kwargs):
        '''
        Calculates the Jones polynomial of the :class:PlanarDiagram using the optimisations
        of the Mathematica package KnotTheory as detailed on the Knot Atlas website

        Will not work on PlanarDiagrams with Points
        '''
        root_of_unity = kwargs.get('root_of_unity')
        if root_of_unity is not None:
            if root_of_unity == -1:
                root_of_unity = 2
            root = n.exp(2 * n.pi * 1.j / root_of_unity)
            root = root.real.round(3) + root.imag.round(3)*1.j
            A = (root + 0j) ** 0.25
        else:
            import sympy as sym
            A = sym.var('A')
            q = sym.var('q')

        crossings = crossings_connectedness_order(self)
        if len(crossings) < 2:
            return 1

        if root_of_unity is None:

            web = []
            crossing = crossings[0]
            crossing = PlanarDiagram(crossing)
            diag_1, diag_2 = kauffman_skein(crossing)
            web.append([diag_1, [1, 0, 1]])
            web.append([diag_2, [1, 0, -1]])


            for crossing in crossings[1:]:
                crossing = PlanarDiagram(crossing)
                diag_1, diag_2 = kauffman_skein(crossing)
                new_web = []

                for term in web:
                    diag_1_product = PlanarDiagram([x for x in diag_1] + [x for x in term[0]]).contract_points()
                    diag_2_product = PlanarDiagram([x for x in diag_2] + [x for x in term[0]]).contract_points()
                    contracted_1 = 0
                    contracted_2 = 0
                    diag_1_new = []
                    diag_2_new = []

                    for point in diag_1_product:
                        if len(point.components()) == 1:
                            contracted_1 += 1
                        else:
                            diag_1_new.append(point)
                    for point in diag_2_product:
                        if len(point.components()) == 1:
                            contracted_2 += 1
                        else:
                            diag_2_new.append(point)

                    new_term_1 = []
                    new_term_2 = []
                    for entry in term[1:]:
                        new_term_1.append([entry[0], entry[1] + contracted_1, entry[2] + 1])
                        new_term_2.append([entry[0], entry[1] + contracted_2, entry[2] - 1])

                    new_term_1 = [sorted(diag_1_new)] + contract_exponents(new_term_1)
                    new_term_2 = [sorted(diag_2_new)] + contract_exponents(new_term_2)

                    new_web.append(new_term_1)
                    new_web.append(new_term_2)

                unique_diags = list(set([tuple(x[0]) for x in new_web]))
                contracted_web = [[list(x)] for x in unique_diags]

                for term in new_web:
                    for diag in contracted_web:
                        if diag[0] == term[0]:
                            diag += term[1:]

                contracted_web_2 = []
                for x in contracted_web:
                    diagram = x[0]
                    exponents = contract_exponents(x[1:])
                    to_append = [diagram] + exponents
                    contracted_web_2.append(to_append)

                web = contracted_web_2

            replacement = -A**2 - A**-2
            jones_precursor = 0
            for term in web[0][1:]:
                jones_precursor += term[0] * replacement**term[1] * A**term[2]
            jones_poly = (-A**3)**self.writhe() * jones_precursor / replacement

            return sym.expand(sym.simplify(jones_poly)).subs(A, q**0.25)
        else:
            web = []
            crossing = crossings[0]
            crossing = PlanarDiagram(crossing)
            diag_1, diag_2 = kauffman_skein(crossing)
            web.append([diag_1, A])
            web.append([diag_2, 1.0/A])
            replacement = -A**2 - A**-2

            for crossing in crossings[1:]:
                crossing = PlanarDiagram(crossing)
                diag_1, diag_2 = kauffman_skein(crossing)
                new_web = []

                for term in web:
                    diag_1_product = PlanarDiagram([x for x in diag_1] + [x for x in term[0]]).contract_points()
                    diag_2_product = PlanarDiagram([x for x in diag_2] + [x for x in term[0]]).contract_points()
                    contracted_1 = 0
                    contracted_2 = 0
                    diag_1_new = []
                    diag_2_new = []

                    for point in diag_1_product:
                        if len(point.components()) == 1:
                            contracted_1 += 1
                        else:
                            diag_1_new.append(point)
                    for point in diag_2_product:
                        if len(point.components()) == 1:
                            contracted_2 += 1
                        else:
                            diag_2_new.append(point)

                    new_term_1 = [sorted(diag_1_new)] + [term[1] * (replacement ** contracted_1) * A]
                    new_term_2 = [sorted(diag_2_new)] + [term[1] * (replacement ** contracted_2) * 1.0/A]

                    new_web.append(new_term_1)
                    new_web.append(new_term_2)

                unique_diags = list(set([tuple(x[0]) for x in new_web]))
                contracted_web = [[list(x)] for x in unique_diags]

                for term in new_web:
                    for diag in contracted_web:
                        if diag[0] == term[0]:
                            if len(diag) == 1:
                                diag.append(term[1])
                            else:
                                diag[1] += term[1]

                web = contracted_web

            jones_precursor = web[0][1]
            jones_poly = (-A**3)**self.writhe() * jones_precursor / replacement

            return round(abs(jones_poly))


class Crossing(list):
    '''
    A single crossing in a planar diagram. Each :class:`PlanarDiagram`
    is a list of these.

    Parameters
    ----------
    a : int or None
        The first entry in the list of lines meeting at this Crossing.
    b : int or None
        The second entry in the list of lines meeting at this Crossing.
    c : int or None
        The third entry in the list of lines meeting at this Crossing.
    d : int or None
        The fourth entry in the list of lines meeting at this Crossing.
    '''

    def __init__(self, a=None, b=None, c=None, d=None):
        super(Crossing, self).__init__()
        self.extend([a, b, c, d])

    def valid(self):
        '''
        True if all intersecting lines are not None.
        '''
        if all([entry is not None for entry in self]):
            return True
        return False

    def components(self):
        '''
        Returns a de-duplicated list of lines intersecting at this Crossing.

        :rtype: list
        '''
        return list(set(self))

    def __str__(self):
        return 'X_{{{0},{1},{2},{3}}}'.format(
            self[0], self[1], self[2], self[3])

    def __repr__(self):
        return self.__str__()

    def as_mathematica(self):
        '''
        Get a string of mathematica code that can represent the Crossing
        in mathematica's knot library.

        The mathematica code won't be valid if any lines of self are None.

        :rtype: str
        '''
        return 'X[{}, {}, {}, {}]'.format(
            self[0], self[1], self[2], self[3])

    def __hash__(self):
        return tuple(self).__hash__()

    def update_line_number(self, old, new):
        '''
        Replaces all instances of the given line number in self.

        Parameters
        ----------
        old : int
            The old line number
        new : int
            The number to replace it with
        '''
        for i in range(4):
            if self[i] == old:
                self[i] = new

    def sign(self):
        check = self[1] - self[3]
        if check == 1 or check < -1:
            return 1
        else:
            return -1


class Point(list):
    '''
    A bivalent vertex in a planar diagram. These are formed when a
    crossing is smoothed and indicate which previously distinct semi-
    arcs are now the same.

    Parameters
    ----------
    a : int or None
        The first entry in the list of lines meeting at this BiVertex.
    b : int or None
        The second entry in the list of lines meeting at this BiVertex.
    '''

    def __init__(self, a=None, b=None):
        super(Point, self).__init__()
        self.extend(sorted([a, b]))

    def valid(self):
        '''
        True if all intersecting lines are not None.
        '''
        if all([entry is not None for entry in self]):
            return True
        return False

    def components(self):
        '''
        Returns a de-duplicated list of lines intersecting at this Crossing.

        :rtype: list
        '''
        return list(set(self))

    def __str__(self):
        return 'P_{{{0},{1}}}'.format(
            self[0], self[1])

    def __repr__(self):
        return self.__str__()

    #def as_mathematica(self):
    #    '''
    #    Get a string of mathematica code that can represent the Crossing
    #    in mathematica's knot library.

    #    The mathematica code won't be valid if any lines of self are None.

    #    :rtype: str
    #    '''
    #    return 'P[{}, {}]'.format(
    #        self[0], self[1])

    def __hash__(self):
        return tuple(self).__hash__()

    def update_line_number(self, old, new):
        '''
        Replaces all instances of the given line number in self.

        Parameters
        ----------
        old : int
            The old line number
        new : int
            The number to replace it with
        '''
        for i in range(2):
            if self[i] == old:
                self[i] = new


def shorthand_to_crossings(s):
    '''
    Takes a planar diagram shorthand string, and returns a list of
    :class:`Crossing`s.
    '''
    crossings = []
    cs = s.split(' ')
    for entry in cs:
        entry = entry.split('_')
        if entry[0] == 'X':
            a, b, c, d = [int(j) for j in entry[1].split(',')]
            crossings.append(Crossing(a, b, c, d))
        elif entry[0] == 'P':
            a, b = [int(j) for j in entry[1].split(',')]
            crossings.append(Point(a, b))
    return crossings


def gausscode_to_crossings(gc):
    cl = gc._gauss_code
    crossings = []
    incomplete_crossings = {}
    line_lengths = [len(line) for line in cl]
    total_lines = sum(line_lengths)
    line_indices = [1] + list(n.cumsum(line_lengths)[:-1] + 1)

    curline = 1
    for i, line in enumerate(cl):
        curline = line_indices[i]
        for index, over, clockwise in line:
            if index in incomplete_crossings:
                crossing = incomplete_crossings.pop(index)
            else:
                crossing = Crossing()

            inline = curline
            curline += 1
            if curline >= (line_indices[i] + line_lengths[i]):
                curline = line_indices[i]
            outline = curline

            if over == -1:
                crossing[0] = inline
                crossing[2] = outline
                crossings.append(crossing)
            else:
                if clockwise == 1:
                    crossing[3] = inline
                    crossing[1] = outline
                else:
                    crossing[1] = inline
                    crossing[3] = outline

            if not crossing.valid():
                incomplete_crossings[index] = crossing

    return crossings

def kauffman_skein(planar_diagram, **kwargs):
    '''
    Performs Kauffman's skein relation on the given planar diagram. For
    a given crossing, this is X_a,b,c,d -> A P_a,d P_b,c + A^-1 P_a,b P_c,d.
    The A's here are ignored.

    The two planar diagrams returned are in the order P_a,d P_b,c then
    P_a,b P_c,d and have their :class:`Point`s contracted according to
    contract_points
    '''
    crossing_components = kwargs.get('crossing_components')
    crossing_number = kwargs.get('crossing_number')
    if crossing_components is not None:
        component = planar_diagram.index(Crossing(crossing_components[0],
                                         crossing_components[1],
                                         crossing_components[2],
                                         crossing_components[3]))
    elif crossing_number is not None:
        pd_crossings = [x for x in planar_diagram if type(x) == Crossing]
        crossing = pd_crossings[crossing_number]
        component = planar_diagram.index(crossing)
    else:
        seeking_crossing = True
        component = 0
        while seeking_crossing:
            if type(planar_diagram[component]) is Crossing:
                seeking_crossing = False
            else:
                component += 1
    import copy
    planar_diagram_1 = copy.copy(planar_diagram)
    planar_diagram_2 = copy.copy(planar_diagram)
    del planar_diagram_1[component]
    del planar_diagram_2[component]

    planar_diagram_1.append(Point(planar_diagram[component][0],
                                  planar_diagram[component][3]))
    planar_diagram_1.append(Point(planar_diagram[component][1],
                                  planar_diagram[component][2]))
    planar_diagram_2.append(Point(planar_diagram[component][0],
                                  planar_diagram[component][1]))
    planar_diagram_2.append(Point(planar_diagram[component][2],
                                  planar_diagram[component][3]))
    if type(planar_diagram) == PlanarDiagram:
        return planar_diagram_1.contract_points(), planar_diagram_2.contract_points()
    else:
        return planar_diagram_1, planar_diagram_2

def kauffman_all_crossings(planar_diagram):
    #pd_crossings = [x for x in planar_diagram if type(x) == Crossing]
    pd_crossings = crossings_connectedness_order(planar_diagram)
    crossing_components = [pd_crossings[0][0], pd_crossings[0][1],
                           pd_crossings[0][2], pd_crossings[0][3]]
    pd_1, pd_2 = (
        kauffman_skein(planar_diagram, crossing_components=crossing_components))
    split_diagram = [[1, pd_1], [-1, pd_2]]
    #print 1, pd_crossings[0]
    i = 2
    for crossing in pd_crossings[1:]:
        new_split_diagram = []
        #print i, crossing
        i += 1
        for diagram in split_diagram:
            coeff_1 = diagram[0] + 1
            coeff_2 = diagram[0] - 1
            crossing_components = [crossing[0], crossing[1],
                                   crossing[2], crossing[3]]
            new_pd_1, new_pd_2 = (
                kauffman_skein(diagram[1], crossing_components=crossing_components))
            new_split_diagram.append([coeff_1, new_pd_1])
            new_split_diagram.append([coeff_2, new_pd_2])
        split_diagram = new_split_diagram
    return split_diagram

def crossings_connectedness_order(planar_diagram):
    crossings = [x for x in planar_diagram if type(x) == Crossing]

    inner_components = crossings[0].components()
    crossings_remaining = crossings[1:]
    ordered_crossings = [crossings[0]]

    for i in range(len(crossings_remaining)):
        connectedness = []
        for crossing in crossings_remaining:
            new_components = (len(list(set(crossing.components() + inner_components))) -
                              len(inner_components))

            connectedness.append(len(crossing.components()) - new_components)
        most_connected_index = connectedness.index(max(connectedness))
        inner_components += crossings_remaining[most_connected_index].components()
        inner_components = list(set(inner_components))
        ordered_crossings.append(crossings_remaining[most_connected_index])
        del crossings_remaining[most_connected_index]

    return ordered_crossings

def contract_points(planar_diagram):
    '''
    For appropriately contracting :class: `Points` in a
    :class: `PlanarDiagram` According to the following rules:

    P_a,b P_b,c -> P_a,c
    P_a,b P_a,b -> P_a,a
    '''
    import copy

    pd_crossings = [x for x in planar_diagram if type(x) == Crossing]
    pd_points = [x for x in planar_diagram if type(x) == Point]

    pd_trimmed = PlanarDiagram()
    for crossing in pd_crossings:
        pd_trimmed.append(crossing)

    substituted_points = []
    for i in range(len(pd_points)):
        if pd_points[i] not in substituted_points:
            pd_trimmed.append(pd_points[i])
            totally_simplified = False
            while not totally_simplified:
                pd_trimmed_old = copy.copy(pd_trimmed)
                for j in range(i+1, len(pd_points)):
                    if (len(pd_trimmed[-1].components()) > 1 and
                            len(pd_points[j].components()) > 1 and
                            pd_points[j] not in substituted_points):
                        # This skips P_a,a situations
                        if pd_trimmed[-1].components() != pd_points[j].components():
                            # This conditions skips (P_a,b P_a,b) situations
                            substituted = False
                            p = 0
                            while substituted is not True and p != 2:
                                q = 0
                                while substituted is not True and q != 2:
                                    if pd_trimmed[-1].components()[p] == pd_points[j].components()[q]:
                                        pd_trimmed.append(Point(pd_trimmed[-1].components()[(p+1) % 2],
                                                                pd_points[j].components()[(q+1) % 2]))
                                        del pd_trimmed[-2]
                                        substituted = True
                                        substituted_points.append(pd_points[j])
                                    q += 1
                                p += 1
                        else:
                            # This deals with (P_a,b P_a,b) situations
                            pd_trimmed.append(Point(pd_trimmed[-1].components()[0],
                                                    pd_trimmed[-1].components()[0]))
                            del pd_trimmed[-2]
                            substituted_points.append(pd_points[j])
                if pd_trimmed_old == pd_trimmed:
                    totally_simplified = True
    return pd_trimmed

def contract_exponents(exponents_array):
    '''
    Takes a set of exponents of the form [[x1, y1, z1], [x2, y2, z2], ...] where,
    given a Jones polynomial state such as (-A^2 - A^-2)^y * A^z * (P_1,2 P_3,4)
    the y's and z's are the exponents shown and x is the number of terms with equal
    y, z and point collection.

    This functions gathers like terms and returns an array of the same form
    '''
    unique_yz = list(set([tuple(x[1:]) for x in exponents_array]))
    unique_yz = [list(x) for x in unique_yz]

    contracted_exponents = []

    for yzs in unique_yz:
        total = 0
        for terms in exponents_array:
            if terms[1:] == yzs:
                total += terms[0]
        contracted_exponents.append([total, yzs[0], yzs[1]])

    return contracted_exponents




'''
GaussCode
=========

Classes for working with Gauss codes representing planar projections
of curves.

See class documentation for more details.

API documentation
~~~~~~~~~~~~~~~~~
'''

from __future__ import print_function
import numpy as n
import re
import sys


class GaussCode(object):
    '''Class for containing and manipulating Gauss codes.

    By default you must pass an extended Gauss code that includes the
    sign of each crossing ('c' for clockwise or 'a' for
    anticlockwise), e.g. ``1+c,2-c,3+c,1-c,2+c,3-c`` for the trefoil
    knot. If you do not know the crossing signs you can instead call
    :meth:`GaussCode.calculating_orientations`, e.g.
    :code:`gc = GaussCode.calculating_orientations('1+,2-,3+,1-,2+,3-')`.

    The length of a Gauss code (e.g. ``len(GaussCode())``) is the
    number of crossings in it.

    Parameters
    ----------
    crossings : array-like or string or PlanarDiagram or GaussCode
        A raw_crossings array from a
        :class:`~pyknotid.spacecurves.spacecurves.Knot`
        or :class:`~pyknotid.spacecurves.spacecurves.Link`, or a string
        representation of the form (e.g.)
        ``1+c,2-c,3+c,1-c,2+c,3-c``, with commas between entries,
        and with multiple link components separated by spaces and/or
        newlines. If a PlanarDiagram or GaussCode is passed, the code
        is duplicated.
    verbose : bool
        Whether to print information during calculations. Defaults to
        True.

    '''

    def __init__(self, crossings='', verbose=True):
        from pyknotid.representations import planardiagram
        if isinstance(crossings, str):
            self._init_from_string(crossings)
        elif isinstance(crossings, GaussCode):
            self._gauss_code = [row.copy() for row in crossings._gauss_code]
        elif isinstance(crossings, planardiagram.PlanarDiagram):
            raise NotImplementedError(
                'planar diagram -> gauss code not implemented')
        else:
            if isinstance(crossings, n.ndarray):
                crossings = [crossings]
            self._init_from_raw_crossings_array(crossings)

        self.crossing_numbers = _get_crossing_numbers(self._gauss_code)
        self.verbose = verbose

    def copy(self):
        return GaussCode(self)

    def mirrored(self):
        '''Returns a copy of self with crossing orientations reversed.'''
        gc = self.copy()
        for row in gc._gauss_code:
            row[:, 2] *= -1
        return gc

    def flipped(self):
        '''Returns a copy of self with crossing over/under switched.'''
        gc = self.copy()
        for row in gc._gauss_code:
            row[:, 1] *= -1
            row[:, 2] *= -1
        return gc

    @classmethod
    def calculating_orientations(cls, code):
        '''Takes a Gauss code without crossing orientations and returns an
        equivalent Gauss code (though not necessarily of the same length).

        This works by generating a space curve and finding its
        self-intersections on projection. This is overkill for the
        problem, but works.

        '''
        code = code.replace(',', 'c,')
        code = code.replace(' ', 'c ')
        code = code.replace('\n', 'c\n')
        code = ''.join([code, 'c'])

        gc = GaussCode(code)
        from pyknotid.representations import representation
        rep = representation.Representation(gc)
        # rep.draw_planar_graph()
        # print('ready to do space curve')
        sp = rep.space_curve()
        # sp.rotate()
        return sp.gauss_code()

    def contains_virtual(self):
        for row in self._gauss_code:
            if n.any(row[:, 0] < 0):
                return True
        return False

    def without_virtual(self):
        '''Returns a version of the Gauss code without explicit virtual
        crossings.'''
        gc = GaussCode(self)

        new_rows = []

        for row in gc._gauss_code:
            keep = n.ones(len(row), dtype=n.bool)
            virtual_cs = n.argwhere(row[:, 0] < 0)
            indices = virtual_cs.flatten()
            for index in indices:
                keep[index] = False
            new_row = row[keep].copy()
            new_rows.append(new_row)

        gc._gauss_code = new_rows
        gc.crossing_numbers = set([c for c in gc.crossing_numbers if c >= 0])
        return gc

    def __len__(self):
        return int(sum([len(c) for c in self._gauss_code]) / 2)

    def _init_from_raw_crossings_array(self, crossings):
        '''
        Takes a list of crossings in (optionally multiple) lines,
        and converts to internal Gauss code representation.
        '''
        assigned_indices = {}
        gauss_code = []
        current_index = 1
        current_virtual_index = -1
        for line in crossings:
            line_gauss_code = []
            for ident, other_ident, over, clockwise in line:
                over = int(over)
                clockwise = int(clockwise)
                if ident not in assigned_indices:
                    if over != 0:
                        assigned_indices[other_ident] = current_index
                        index = current_index
                        current_index += 1
                    else:
                        assigned_indices[other_ident] = current_virtual_index
                        index = current_virtual_index
                        current_virtual_index -= 1
                else:
                    index = assigned_indices.pop(ident)
                line_gauss_code.append([index, over, clockwise])
            gauss_code.append(n.array(line_gauss_code))

        self._gauss_code = gauss_code

    def _init_from_string(self, crossings):
        '''
        Converts the string into internal Gauss code representation.
        '''
        regex = re.compile('[ \n]*')
        lines = regex.split(crossings)

        gauss_code = []
        over_under = {'+': 1, '-': -1, 'v': 0}
        signs = {'c': 1, 'a': -1, 'v': 1}

        for line in lines:
            line_gauss_code = []
            line_crossings = line.split(',')
            for line_crossing in line_crossings:
                line_gauss_code.append([int(line_crossing[:-2]),
                                        over_under[line_crossing[-2]],
                                        signs[line_crossing[-1]]])
            gauss_code.append(n.array(line_gauss_code))

        self._gauss_code = gauss_code

    def __repr__(self):
        out_strs = []

        gauss_code = self._gauss_code
        for line in gauss_code:
            if len(line) == 0:
                out_strs.append('----')
                out_strs.append(',')
            else:
                for entry in line:
                    out_strs.append(str(entry[0]))
                    classical = entry[1] != 0
                    if classical:
                        over = (entry[1] > 0)
                        if over:
                            out_strs.append('+')
                        else:
                            out_strs.append('-')

                        clockwise = (entry[2] > 0)
                        if clockwise:
                            out_strs.append('c')
                        else:
                            out_strs.append('a')
                    else:
                        out_strs.append('v')

                    out_strs.append(',')
            out_strs = out_strs[:-1]
            out_strs.append('\n')
        out_strs = out_strs[:-1]
        return ''.join(out_strs)

    def __str__(self):
        return repr(self)

    def _do_reidemeister_moves(self, one=True, two=True, one_extended=True):
        '''
        Performs the given Reidemeister moves a single time, iterating
        over all the crossings of self.
        '''

        code = self._gauss_code
        crossing_numbers = self.crossing_numbers

        rm2_store = {}

        # These lists keep track of which crossings have been removed, to
        # avoid modifying the arrays every time an RM is performed
        keeps = [n.ones(l.shape[0], dtype=bool) for l in code]

        # First do RM1 and RM2
        for line_index, line in enumerate(code):
            keep = keeps[line_index]

            for row_index, row in enumerate(line):
                next_index = (row_index + 1) % len(line)
                next_row = line[next_index]

                if (one and (row[0] == next_row[0]) and keep[row_index] and
                    keep[next_index]):
                    number = row[0]
                    crossing_numbers.remove(number)
                    keep[row_index] = False
                    keep[next_index] = False

                if (two and keep[row_index] and keep[next_index] and
                    (row[1] == next_row[1])):  # both over or under
                    numbers = tuple(sorted([row[0], next_row[0]]))
                    if numbers not in rm2_store:
                        rm2_store[numbers] = (line_index, row_index,
                                              next_index)
                    else:
                        other_indices = rm2_store.pop(numbers)
                        crossing_numbers.remove(row[0])
                        crossing_numbers.remove(next_row[0])
                        keep[row_index] = False
                        keep[next_index] = False
                        keeps[other_indices[0]][other_indices[1]] = False
                        keeps[other_indices[0]][other_indices[2]] = False

        # Next do RM1_extended separately (could be more efficient?)
        if one_extended:
            # Do extended RM1 as a separate step
            print
            code = [(line[keep] if len(line) > 0 else line) for (line, keep) in zip(code, keeps)]
            keeps = [n.ones(l.shape[0], dtype=bool) for l in code]

            crossing_indices = {}
            for line_index, line in enumerate(code):
                for row_index, row in enumerate(line):
                    identifier, over, crossings = row
                    if identifier not in crossing_indices:
                        crossing_indices[identifier] = []
                    crossing_indices[identifier].append((line_index,
                                                         row_index))
            for number in list(crossing_numbers):
                if number not in crossing_numbers:
                    continue  # The crossing has already been removed
                locations = crossing_indices[number]

                if locations[0][0] != locations[1][0]:  # not on same line
                    continue
                line_index = locations[0][0]
                first_index = locations[0][1]
                second_index = locations[1][1]
                first_index, second_index = sorted(
                    [first_index, second_index])

                # First, check crossings in the middle of the list
                in_between = code[line_index][first_index+1:second_index]
                in_between_keeps = keeps[line_index][first_index+1:second_index]
                if len(in_between) > 0:
                    in_between = in_between[in_between_keeps]

                if n.abs(n.sum(in_between[:, 1])) == len(in_between):
                    # all crossings over or under
                    keeps[line_index][first_index] = False
                    keeps[line_index][second_index] = False
                    for entry in in_between:
                        identifier = entry[0]
                        if identifier in crossing_numbers:
                            crossing_numbers.remove(identifier)
                        indices = crossing_indices[identifier]
                        keeps[indices[0][0]][indices[0][1]] = False
                        keeps[indices[1][0]][indices[1][1]] = False
                    crossing_numbers.remove(number)

                # Second, wrap around the list if a big rm1 wasn't performed
                if number not in crossing_numbers:
                    continue
                in_between = n.vstack((code[line_index][second_index+1:],
                                       code[line_index][:first_index]))
                in_between_keeps = n.hstack((keeps[line_index][second_index+1:],
                                             keeps[line_index][:first_index]))
                if len(in_between) > 0:
                    in_between = in_between[in_between_keeps]

                if n.abs(n.sum(in_between[:, 1])) == len(in_between):
                    keeps[line_index][first_index] = False
                    keeps[line_index][second_index] = False
                    for entry in in_between:
                        identifier = entry[0]
                        if identifier in crossing_numbers:
                            crossing_numbers.remove(identifier)
                        indices = crossing_indices[identifier]
                        keeps[indices[0][0]][indices[0][1]] = False
                        keeps[indices[1][0]][indices[1][1]] = False
                    crossing_numbers.remove(number)


        # Get rid of all crossings that have been removed by RMs
        self._gauss_code = [(line[keep] if len(line) > 0 else line) for (line, keep) in zip(code, keeps)]
        self.crossing_numbers = crossing_numbers

    def simplify(self, one=True, two=True, one_extended=True):
        '''
        Simplifies the GaussCode, performing the given Reidemeister moves
        everywhere possible, as many times as possible, until the
        GaussCode is no longer changing.

        This modifies the GaussCode - (non-topological) information may
        be lost!

        Parameters
        ----------
        one : bool
            Whether to use Reidemeister 1, defaults to True.
        two : bool
            Whether to use Reidemeister 2, defaults to True.
        one_extended : bool
            Whether to use extended Reidemeister 1, which removes crossings
            connected by arcs which include only over or only under crossings
            (and which must thus be topologically irrelevant). Defaults
            to True.
        '''

        if self.verbose:
            print('Simplifying: initially {} crossings'.format(
                n.sum([len(line) for line in self._gauss_code])))

        number_of_runs = 0
        while True:
            original_gc = self._gauss_code
            original_len = n.sum([len(line) for line in original_gc])
            self._do_reidemeister_moves(one, two)
            new_gc = self._gauss_code
            new_len = n.sum([len(line) for line in new_gc])
            number_of_runs += 1
            if self.verbose:
                sys.stdout.write('\r-> {} crossings after {} runs'.format(
                    n.sum([len(line) for line in new_gc]), number_of_runs))
                sys.stdout.flush()
            if new_len == original_len:
                break

        if self.verbose:
            print()

    def reindex_crossings(self):
        '''Replaces the indices of the crossings in the Gauss code with the
        integers from 1 to its length.

        Note that this modifies the Gauss code in place, the previous indices
        are not recorded.
        '''
        num_crossings = len(self)

        new_number = 1
        numbers_dict = {}

        for line in self._gauss_code:
            for i, row in enumerate(line):
                number = row[0]
                if number in numbers_dict:
                    row[0] = numbers_dict.pop(number)
                else:
                    numbers_dict[row[0]] = new_number
                    row[0] = new_number
                    new_number += 1

        self.crossing_numbers = set(range(1, num_crossings + 1))

    def _remove_crossing(self, number):
        gc = self._gauss_code
        new_gc = []
        for row in gc:
            new_gc.append(row[row[:, 0] != number])
        self._gauss_code = new_gc
        self.crossing_numbers = _get_crossing_numbers(self._gauss_code)


def _get_crossing_numbers(gc):
    '''
    Given GaussCode internal data, returns a list of all
    the crossing numbers within
    '''
    crossing_vals = set()
    for line in gc:
        for entry in line:
            crossing_vals.add(entry[0])
    return crossing_vals




from __future__ import print_function, division

import numpy as n

from pyknotid.utils import ensure_shape_tuple, vprint
from periodiccell import _cram_into_cell, _cut_line_at_jumps, _interpret_line

from collections import defaultdict

try:
    from pyknotid.spacecurves import chelpers
except ImportError:
    from pyknotid.spacecurves import helpers as chelpers

ROTATION_MAGIC_NUMBERS = (0.02, 1.1)
ROTATION_MAGIC_NUMBERS = (0.53, 1.0)

class PeriodicKnot(object):
    def __init__(self, points, period_vector=None, autorotate=True, repeats=1):

        if repeats < 1:
            raise ValueError('Must be at least 1 repeat of the base points')
        if repeats > 1:
            k = PeriodicKnot(points, period_vector=period_vector, autorotate=False)
            new_points = [points]
            for i in range(1, repeats):
                new_points.append(k.translated_points(i))
            points = n.vstack(new_points)
        
        self.points = n.array(points).copy()
        self._period_vector = period_vector

        from pyknotid.spacecurves.rotation import rotate_to_top
        if autorotate:
            self._apply_matrix(rotate_to_top(*ROTATION_MAGIC_NUMBERS))

        if self._period_vector is not None:
            self._period_vector = rotate_to_top(*ROTATION_MAGIC_NUMBERS).dot(self._period_vector)

    @classmethod
    def from_periodic_line(cls, line, shape, perturb=True, **kwargs):
        '''Returns a :class:`PeriodicKnot` instance in which the line has been
        unwrapped through
        the periodic boundaries.

        Parameters
        ----------
        line : array-like
            The Nx3 vector of points in the line
        shape : array-like
            The x, y, z distances of the periodic boundary
        perturb : bool
            If True, translates and rotates the knot to avoid any lattice
            problems.
        '''
        shape = ensure_shape_tuple(shape)

        points = line.copy()
        points += n.array([0.00123, 0.00231, 0.00321])

        dx, dy, dz = shape
        for i in range(1, len(points)):
            prevLine = points[i-1]
            curLine = points[i]
            rest = points[i:]
            change = curLine - prevLine
            if -1.05*dx < change[0] < -0.95*dx:
                rest[:, 0] += dx
            if 1.05*dx > change[0] > 0.95*dx:
                rest[:, 0] -= dx
            if -1.05*dy < change[1] < -0.95*dy:
                rest[:, 1] += dy
            if 1.05*dy > change[1] > 0.95*dy:
                rest[:, 1] -= dy
            if -1.05*dz < change[2] < -0.95*dz:
                rest[:, 2] += dz
            if 1.05*dz > change[2] > 0.95*dz:
                rest[:, 2] -= dz

        return cls(points)



    def __len__(self):
        return len(self.points)

    @property
    def period_vector(self):
        if self._period_vector is not None:
            return n.array(self._period_vector)
        return self.points[-1] - self.points[0]

    def rotate_period_vector_to_x(self):
        v = self.period_vector
        from pyknotid.spacecurves.rotation import (
            rotate_vector_to_top, rotate_axis_angle)
        self._apply_matrix(rotate_vector_to_top(v))
        self._apply_matrix(rotate_axis_angle((0, 1., 0), n.pi/2.))

    def get_periodic_segments(self):
        '''Returns a list of line segments making up the curve, cut at the
        edges of a single periodic repeat.
        '''
        

    def interpolate(self, factor=2):
        if factor == 1:
            return
        new_points = []
        for i in range(len(self.points) - 1):
            v1 = self.points[i]
            v2 = self.points[i+1]
            interpolated_points = n.zeros((factor, 3))
            interpolated_points[:, 0] = n.linspace(v1[0], v2[0], factor + 1)[:-1]
            interpolated_points[:, 1] = n.linspace(v1[1], v2[1], factor + 1)[:-1]
            interpolated_points[:, 2] = n.linspace(v1[2], v2[2], factor + 1)[:-1]
            new_points.append(interpolated_points)
        self.points = n.vstack(new_points)

    def nearest_non_overlapping_translation(self):
        '''Returns the minimal number of translations that can be used in
        raw_crossings to ensure no crossings are missed.'''

        points = self.points[:, :2]
        end_end_distance = mag(points[-1] - self.translated_points(1)[-1, :2])

        max_dist_from_start = n.max(n.apply_along_axis(mag, 1, points - points[0]))
        max_dist_from_end = n.max(n.apply_along_axis(mag, 1, points - points[-1]))
        largest_possible_distance = n.max([max_dist_from_start, max_dist_from_end])

        # print('largest possible distance', largest_possible_distance)
        # print('end_end', end_end_distance)

        return int(n.ceil(largest_possible_distance / end_end_distance) + 4)

    def roll(self, num):
        pv = self.period_vector
        ps = self.points

        for _ in range(abs(num)):
            if num > 0:
                ps = n.vstack((ps[1:], [ps[1] + pv]))
            elif num < 0:
                ps = n.vstack(([ps[-2] - pv], ps[:-1]))
        self.points = ps

    def plot_with_translation(self, with_translation=0, **kwargs):
        from pyknotid.spacecurves import OpenKnot, Link
        if with_translation == 0:
            OpenKnot(self.points).plot(**kwargs)
        else:
            Link([self.points,
                  self.translated_points(with_translation)]).plot(
                      colours=['red', 'blue'])

    def plot(self, num_translations=3, colour_minimal_unfoldings=True,
             **kwargs):
        from pyknotid.spacecurves import OpenKnot, Link
        if not colour_minimal_unfoldings:
            points = self.points_with_translations(num_translations)
            k = OpenKnot(points)
            k.plot(**kwargs)
            return k
        lines = []
        for i in range(-num_translations, num_translations+1):
            lines.append(self.translated_points(i))
        l = Link(lines)
        l.plot(closed=False, **kwargs)
        return l

    def translated_points(self, num):
        return self.points + num * self.period_vector

    def smooth(self, **kwargs):
        '''Smooths each of the x, y, z components of self.points, using the
        smooth function of
        :class:`pyknotid.spacecurves.spacecurve.Spacecurve`.

        .. warning:: This *can* change the topology of the curve.
        '''
        from pyknotid.spacecurves.openknot import OpenKnot
        k = OpenKnot(self.points)
        k.smooth(**kwargs)
        self.points = k.points

    def plot_projection_with(self, num_translations=0, mat=None):
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        points = self.points
        translated_points = self.translated_points(num_translations)
        if mat is not None:
            points = n.apply_along_axis(mat.dot, 1, points)
            translated_points = n.apply_along_axis(
                mat.dot, 1, translated_points)
        ax.plot(points[:, 0], points[:, 1], color='purple', linewidth=1.5)
        ax.plot(translated_points[:, 0],
                translated_points[:, 1], color='green')
        fig.show()
        return fig, ax

    def points_with_translations(self, num_translations=3):
        if num_translations == 0:
            return self.points
        new_points = []
        for i in range(-num_translations, num_translations):
            new_points.append(self.translated_points(i)[:-1])
        new_points.append(self.translated_points(num_translations))
        points = n.vstack(new_points)
        return points

    def rotate(self, angles=None):

        '''
        Rotates all the points of self by the given angles in each axis.

        Parameters
        ----------
        angles : array-like
            The rotation angles about x, y and z. If None, random
            angles are used. Defaults to None.
        '''
        from pyknotid.utils import get_rotation_matrix
        if angles is None:
            angles = n.random.random(3) * 2*n.pi
        phi, theta, psi = angles
        rot_mat = get_rotation_matrix(angles)
        self._apply_matrix(rot_mat)

        if self._period_vector is not None:
            self._period_vector = rot_mat.dot(self._period_vector)

    def _apply_matrix(self, mat):
        '''
        Applies the given matrix to all of self.points.
        '''
        self.points = n.apply_along_axis(mat.dot, 1, self.points)

    def plot_projection(self, num_translations=None):
        if num_translations is None:
            num_translations = self.nearest_non_overlapping_translation()
        from pyknotid.visualise import plot_line, plot_projection
        import matplotlib.pyplot as plt
        points = self.points_with_translations(num_translations)
        crossings, core_num, core_index = self.raw_crossings(num_translations)
        plot_crossings = []
        for crossing in crossings:
            x, y, over, orientation = crossing
            xint = int(x)
            r = points[xint]
            dr = points[(xint+1) % len(points)] - r
            plot_crossings.append(r + (x-xint) * dr)
        fig, ax = plt.subplots()
        # prev_points = self.translated_points(-1)
        # ax.plot(prev_points[:, 0], prev_points[:, 1], color='orange', linewidth=1.5)
        fig, ax = plot_projection(points,
                                  crossings=n.array(plot_crossings),
                                  mark_start=True,
                                  fig_ax = (fig, ax),
                                  show=True)
        core_points = points[num_translations * (len(self.points) - 1):
                             num_translations * (len(self.points) - 1) + len(self.points)]
        # ax.plot(points[:, 0], points[:, 1], 'o', color='green')
        ax.plot(core_points[:, 0], core_points[:, 1], color='purple', linewidth=1.5)
        return fig, ax

    def raw_crossings(self, num_translations=None):
        if num_translations is None:
            num_translations = self.nearest_non_overlapping_translation()
        from pyknotid.spacecurves.rotation import rotate_to_top

        points = self.points_with_translations(num_translations)

        num_points = len(points)
        core_num = len(self.points) - 1
        core_index = num_translations * core_num 

        segment_lengths = n.roll(points[:, :2], -1, axis=0) - points[:, :2]
        segment_lengths = n.sqrt(n.sum(segment_lengths**2, axis=1))
        max_segment_length = n.max(segment_lengths[:-1])
        jump_mode = 2

        core_points = points[(core_index):(core_index + core_num + 1)]
        core_segment_lengths = segment_lengths[core_index:(core_index + core_num + 1)]

        first_block = points[:(core_index + 1)]
        first_block_lengths = segment_lengths[:(core_index + 1)]

        second_block = points[(core_index + core_num):]
        second_block_lengths = segment_lengths[(core_index + core_num):]
        
        crossings = []
        for i in range(len(core_points) - 1):
            v0 = core_points[i]
            dv = core_points[i+1] - v0

            vnum = core_index + i

            compnum = 0
            crossings.extend(chelpers.find_crossings(
                v0, dv, first_block, first_block_lengths,
                vnum, compnum,
                max_segment_length,
                jump_mode))

            compnum = core_index + core_num
            crossings.extend(chelpers.find_crossings(
                v0, dv, second_block, second_block_lengths,
                vnum, compnum,
                max_segment_length,
                jump_mode))

            compnum = core_index + i + 2
            crossings.extend(chelpers.find_crossings(
                v0, dv, core_points[i+2:], core_segment_lengths[i+2:],
                vnum, compnum,
                max_segment_length,
                jump_mode))

        crossings.sort(key=lambda j: j[0])
        return n.array(crossings), core_num, core_index

    def alternative_raw_crossings(self, num_translations=None):
        if num_translations is None:
            num_translations = self.nearest_non_overlapping_translation()
        from pyknotid.spacecurves.rotation import rotate_to_top

        points = self.points_with_translations(num_translations)

        from pyknotid.spacecurves.openknot import OpenKnot

        k = OpenKnot(points)

        crossings = k.raw_crossings()

        num_points = len(points)
        core_num = len(self.points) - 1
        core_index = num_translations * core_num 

        return crossings, core_num, core_index

    def alternative_gauss_code(self, num_translations=None):
        if num_translations is None:
            num_translations = self.nearest_non_overlapping_translation()
        crossings, core_num, core_index = self.alternative_raw_crossings(num_translations)

        equivalent_crossing_indices = get_equivalent_crossing_indices(
            crossings, len(self.points) - 1)

        from pyknotid.representations import GaussCode
        gc = GaussCode(crossings)
        # if num_translations == 0:
        #     return gc, defaultdict(set)

        gc_array = gc._gauss_code[0]

        equivalent_crossing_numbers = get_equivalent_crossing_numbers(
            equivalent_crossing_indices, gc._gauss_code[0])

        core_crossings = set()
        multiplicities = defaultdict(lambda *j: 1)
        for raw_crossing, gc_crossing in zip(crossings, gc_array):
            if ((core_index <= raw_crossing[0] <= core_index + core_num) or
                (core_index <= raw_crossing[1] <= core_index + core_num)):
                core_crossings.add(gc_crossing[0])
                multiplicities[gc_crossing[0]] = 1
            if ((core_index <= raw_crossing[0] <= core_index + core_num) and
                (core_index <= raw_crossing[1] <= core_index + core_num)):
                multiplicities[gc_crossing[0]] = 2

        true_crossing_numbers = get_true_crossing_numbers(equivalent_crossing_numbers, core_crossings)

        return gc, core_crossings, multiplicities, true_crossing_numbers

    def gauss_code(self, num_translations=None):
        if num_translations is None:
            num_translations = self.nearest_non_overlapping_translation()
        crossings, core_num, core_index = self.raw_crossings(num_translations)

        equivalent_crossing_indices = get_equivalent_crossing_indices(
            crossings, len(self.points) - 1)

        from pyknotid.representations import GaussCode
        gc = GaussCode(crossings)
        if num_translations == 0:
            return gc, defaultdict(set)

        equivalent_crossing_numbers = get_equivalent_crossing_numbers(
            equivalent_crossing_indices, gc._gauss_code[0])

        translation_numbers = {}
        centre_start = core_index
        code = gc._gauss_code[0]
        for i, row in enumerate(crossings):
            crossing_number = code[i][0]
            if crossing_number not in translation_numbers:
                position1 = row[0]
                translation1 = int(position1 / (len(self.points) - 1)) - num_translations

                position2 = row[1]

                translation2 = int(position2 / (len(self.points) - 1)) - num_translations


                translation_numbers[crossing_number] = (translation1, translation2)

        return gc, equivalent_crossing_numbers, translation_numbers

    def vassiliev_degree_2(self, num_translations=None):
        if num_translations is None:
            num_translations = self.nearest_non_overlapping_translation()
        gc, equivalencies, translations = self.gauss_code(num_translations)
        return periodic_vassiliev_degree_2_without_double_count(gc, equivalencies, translations)

    def alternative_vassiliev_degree_2(self, num_translations=None):
        if num_translations is None:
            num_translations = self.nearest_non_overlapping_translation()
        gc, core_crossings, multiplicities, true_crossing_numbers = self.alternative_gauss_code(num_translations)
        return alternative_periodic_vassiliev_degree_2_without_double_count(gc, core_crossings, multiplicities,
                                                                            true_crossing_numbers)

    def vassiliev_degree_2s(self, number_of_samples=10):
        v2s = []

        from pyknotid.spacecurves.rotation import get_rotation_angles, rotate_to_top            
        angles = get_rotation_angles(number_of_samples)

        for i, angs in enumerate(angles):
            k = PeriodicKnot(self.points.copy())
            k._apply_matrix(rotate_to_top(*angs))
            v2s.append([angs[0], angs[1], k.vassiliev_degree_2()])

        return n.array(v2s)

    def vassiliev_degree_3(self, num_translations=None):
        if num_translations is None:
            num_translations = self.nearest_non_overlapping_translation()
        gc, equivalencies, translations = self.gauss_code(num_translations)
        return periodic_vassiliev_degree_3_without_double_count(gc, equivalencies, translations)

    def alternative_vassiliev_degree_3(self, num_translations=None):
        if num_translations is None:
            num_translations = self.nearest_non_overlapping_translation()
        gc, core_crossings, multiplicities, true_crossing_numbers = self.alternative_gauss_code(num_translations)
        return alternative_periodic_vassiliev_degree_3_without_double_count(gc, core_crossings, 
                                                                            true_crossing_numbers)

    def silly_v3(self, nt=None):
        v3_1 = self.alternative_vassiliev_degree_3(nt)
        self.points[:, 1:] *= -1
        v3_2 = self.alternative_vassiliev_degree_3(nt)
        self.points[:, 1:] *= -1
        return v3_1, v3_2
        if num_translations is None:
            num_translations = self.nearest_non_overlapping_translation()
        gc, core_crossings, multiplicities, true_crossing_numbers = self.alternative_gauss_code(num_translations)
        return alternative_periodic_vassiliev_degree_3_without_double_count(gc, core_crossings, 
                                                                            true_crossing_numbers)
    def vass_flip(self):
        self.points[:, 1:] *= -1

    def alternative_vassiliev_degree_4_conway_z4(self, num_translations=None):
        if num_translations is None:
            num_translations = self.nearest_non_overlapping_translation()
        gc, core_crossings, multiplicities, true_crossing_numbers = self.alternative_gauss_code(num_translations)
        return alternative_periodic_vassiliev_degree_4_conway_z4(gc, core_crossings, 
                                                                            true_crossing_numbers)


    def vassiliev_degree_3s(self, number_of_samples=10):
        v3s = []

        from pyknotid.spacecurves.rotation import get_rotation_angles, rotate_to_top            
        angles = get_rotation_angles(number_of_samples)

        for i, angs in enumerate(angles):
            k = PeriodicKnot(self.points.copy())
            k._apply_matrix(rotate_to_top(*angs))
            v3s.append([angs[0], angs[1], k.vassiliev_degree_3()])

        return n.array(v3s)

    def vassiliev_degree_2_integral(self, num_translations=3):
        ps = self.points_with_translations(num_translations)

        core_num = len(self.points) - 1
        core_index = num_translations * core_num 
        core_end_num = core_index + core_num

        total = 0.
        for i in range(len(ps) - 1):
            vi = ps[i]
            dvi = ps[i+1] - ps[i]
            for j in range(i+1, len(ps) - 1):
                print('ij', i, j)
                vj = ps[j]
                dvj = ps[j+1] - ps[j]
                for k in range(j+1, len(ps) - 1):
                    vk = ps[k]
                    dvk = ps[k+1] - ps[k]
                    for l in range(k+1, len(ps) - 1):
                        vl = ps[l]
                        dvl = ps[l+1] - ps[l]

                        if (not (core_num < i < core_end_num or core_num < k < core_end_num) or
                            not (core_num < j < core_end_num or core_num < l < core_end_num)):
                            continue
                        total += (writhe_contribution(vi, dvi, vk, dvk) *
                                  writhe_contribution(vj, dvj, vl, dvl))
        return total


def mag(v):
    return n.sqrt(v.dot(v))

prefactor = 1 / (4*n.pi)
def writhe_contribution(v0, dv0, v1, dv1):
    diff = v1 - v0
    return (mag(dv0) * mag(dv1) * prefactor *
            n.cross(dv0, dv1).dot(diff) /
            (mag(diff)**3))
    

def get_equivalent_crossing_indices(crossings, span):
    equivalencies = defaultdict(set)
    for i, row1 in enumerate(crossings):
        for j, row2 in enumerate(crossings[i+1:], i+1):
            proximity = n.abs(row1[0] - row2[0]) % span
            if proximity < 0.00000001 or proximity > (span - 0.00000001):
            # if ((n.abs((row1[0] - row2[0])) % span < 0.000001) or
            #     (span - (n.abs((row1[0] - row2[0])) % span) > (span - 0.000001))):
                equivalencies[i].add(j)
                for val in equivalencies[i]:
                    equivalencies[val].add(j)
                    equivalencies[j].add(val)
                equivalencies[j].add(i)
    for key, value in equivalencies.items():
        try:
            value.remove(key)
        except KeyError:
            pass
    return equivalencies


def get_equivalent_crossing_numbers(indices, gauss_code):
    equivalent_crossing_numbers = defaultdict(set)
    for key, values in indices.items():
        number = gauss_code[key][0]
        equivalent_crossing_numbers[number] = set([int(gauss_code[value][0]) for value in values])
        equivalent_crossing_numbers[number].add(number)
    return equivalent_crossing_numbers


class CellKnot(object):
    def __init__(self, points, shape, period=None):
        # points will automatically be folded
        points = _interpret_line(points)

        self.shape = n.array(ensure_shape_tuple(shape))
        self.points = _cut_line_at_jumps(points, self.shape)

        if period is None:
            period = (n.round((self.unfolded_points()[-1] - self.points[0][0]) /
                              self.shape)) 
        self.period = period
    
    @classmethod
    def folding(cls, points, shape, origin=(0, 0, 0.), **kwargs):
        '''Return an instance of PeriodicKnot resulting from folding the
        points within the given shape.'''
        origin = n.array(origin)
        shape = ensure_shape_tuple(shape)

        points -= origin
        points[:, 0] %= shape[0]
        points[:, 1] %= shape[1]
        points[:, 2] %= shape[2]
        return cls(points, shape, **kwargs)

    def plot(self, boundary=True, clf=True, tube_radius=1.0,
             **kwargs):
        from pyknotid.visualise import plot_cell
        boundary = self.shape if boundary else None
        plot_cell([self.points], boundary=boundary, clf=clf,
                  tube_radius=tube_radius, **kwargs)

    def unfolded_points(self):
        points = [self.points[0]]
        current_shift = n.array([0., 0., 0.])
        for i, segment in enumerate(self.points[1:], 1):
            segment = segment - current_shift * self.shape
            diff = segment[0] - points[-1][-1]
            diff = n.round(diff / self.shape)
            current_shift += diff 
            points.append(segment - diff * self.shape)
        return n.vstack(points)

    def unfolded_points_with_translations(self, num=3, mat=None):
        points = []
        unfolded = self.unfolded_points()
        for i in range(-num, num+1):
            points.append(unfolded + self.period_vector() * i)
        points = n.vstack(points)
        if mat is not None:
             points = n.apply_along_axis(mat.dot, 1, points)
        return n.vstack(points)


    def translate(self, distance):
        distance = n.array(ensure_shape_tuple(distance))
        points = self.unfolded_points() + distance
        points = _fold(points, self.shape)
        self.points = _cut_line_at_jumps(points, self.shape)
        self._fix_endpoints()
 
    def _fix_endpoints(self):
        if len(self.points) < 2:
            return
        closing_vector = self.points[-1][-1] - self.points[0][0]
        if n.all(closing_vector < 0.9*self.shape):
            first = self.points[0]
            self.points = self.points[1:]
            last = self.points.pop()
            self.points.append(n.vstack((last, first)))
            
    def period_vector(self):
        return self.period * self.shape

    def raw_crossings_by_unfolding(self, num_translations=0, shift=0):
        from pyknotid.spacecurves.rotation import rotate_to_top
        points = self.unfolded_points_with_translations(
            num_translations, rotate_to_top(*ROTATION_MAGIC_NUMBERS))

        num_points = len(points)
        core_num = len(self.unfolded_points())
        core_index = num_translations * core_num + shift

        segment_lengths = n.roll(points[:, :2], -1, axis=0) - points[:, :2]
        segment_lengths = n.sqrt(n.sum(segment_lengths**2, axis=1))
        max_segment_length = n.max(segment_lengths[:-1])
        jump_mode = 2

        core_points = points[(core_index):(core_index + core_num + 1)]
        core_segment_lengths = segment_lengths[(core_index + shift):(core_index + shift + core_num + 1)]

        first_block = points[:(core_index + 1)]
        first_block_lengths = segment_lengths[:(core_index + 1)]

        second_block = points[(core_index + core_num):]
        second_block_lengths = segment_lengths[(core_index + core_num):]
        
        crossings = []
        for i in range(len(core_points) - 1):
            v0 = core_points[i]
            dv = core_points[i+1] - v0

            vnum = core_index + i

            compnum = 0
            crossings.extend(chelpers.find_crossings(
                v0, dv, first_block, first_block_lengths,
                vnum, compnum,
                max_segment_length,
                jump_mode))

            compnum = core_index + core_num
            crossings.extend(chelpers.find_crossings(
                v0, dv, second_block, second_block_lengths,
                vnum, compnum,
                max_segment_length,
                jump_mode))

            compnum = core_index + i + 2
            crossings.extend(chelpers.find_crossings(
                v0, dv, core_points[i+2:], core_segment_lengths[i+2:],
                vnum, compnum,
                max_segment_length,
                jump_mode))

        crossings.sort(key=lambda j: j[0])
        return n.array(crossings), core_num, core_index

    def plot_projection_by_unfolding(self, num_translations=0, shift=0):
        from pyknotid.visualise import plot_line, plot_projection
        from pyknotid.spacecurves.rotation import rotate_to_top
        points = self.unfolded_points_with_translations(
            num_translations, rotate_to_top(*ROTATION_MAGIC_NUMBERS))
        crossings = self.raw_crossings_by_unfolding(num_translations, shift=shift)[0]
        plot_crossings = []
        for crossing in crossings:
            x, y, over, orientation = crossing
            xint = int(x)
            r = points[xint]
            dr = points[(xint+1) % len(points)] - r
            plot_crossings.append(r + (x-xint) * dr)
        fig, ax = plot_projection(points,
                                  crossings=n.array(plot_crossings),
                                  mark_start=True,
                                  show=True)
        core_points = points[num_translations * len(self.unfolded_points()) + shift:
                             num_translations * len(self.unfolded_points()) + shift + len(self.unfolded_points())]
        # ax.plot(points[:, 0], points[:, 1], 'o', color='green')
        ax.plot(core_points[:, 0], core_points[:, 1], color='purple', linewidth=1.5)
        return fig, ax

    def gauss_code_by_unfolding(self, num_translations=0, mat=None, shift=None):
        '''Returns the Gauss code alongside a list of  that are
        crossings of the non-translated curve with the translated one.'''
        crossings, core_num, core_index = self.raw_crossings_by_unfolding(
            num_translations=num_translations, shift=shift)

        equivalent_crossing_indices = defaultdict(set)
        length = len(self.points)
        for i, row1 in enumerate(crossings):
            for j, row2 in enumerate(crossings[i+1:], i+1):
                # print(i, j, n.abs((row1[0] - row2[0]) % 40))
                if n.abs((row1[0] - row2[0]) % len(self.unfolded_points())) < 0.000001:
                    equivalent_crossing_indices[i].add(j)
                    for val in equivalent_crossing_indices[i]:
                        equivalent_crossing_indices[val].add(j)
                        equivalent_crossing_indices[j].add(val)
                    equivalent_crossing_indices[j].add(i)
        for key, value in equivalent_crossing_indices.items():
            try:
                value.remove(key)
            except KeyError:
                pass
        
        from pyknotid.representations import GaussCode
        gc = GaussCode(crossings)
        if num_translations == 0:
            return gc, defaultdict(set)

        equivalent_crossing_numbers = defaultdict(set)
        code = gc._gauss_code[0]
        for key, values in equivalent_crossing_indices.items():
            number = code[key][0]
            equivalent_crossing_numbers[number] = set([int(code[value][0]) for value in values])
            equivalent_crossing_numbers[number].add(number)

        translation_numbers = {}
        if shift is None:
            shift = 0
        centre_start = num_translations * len(self.unfolded_points()) + shift
        for i, row in enumerate(crossings):
            crossing_number = code[i][0]
            if crossing_number not in translation_numbers:
                position1 = row[0]
                translation1 = int(n.floor((position1 - centre_start) / (len(self.unfolded_points() + 0))))

                position2 = row[1]
                translation2 = int(n.floor((position2 - centre_start) / (len(self.unfolded_points() + 0))))

                translation_numbers[crossing_number] = (translation1, translation2)
                
        return gc, equivalent_crossing_numbers, translation_numbers
        

    def raw_crossings(self, num_translations=0, mat=None, shift=0):
        from pyknotid.spacecurves import OpenKnot
        points = self.unfolded_points()
        points = n.vstack((points[shift:], points[:(shift+1)] + self.period_vector()))
        translated_points = points + num_translations * self.period_vector()

        if mat is not None:
            points = n.apply_along_axis(mat.dot, 1, points)
            translated_points = n.apply_along_axis(mat.dot, 1, translated_points)

        # 1) get crossings of self with self
        self_crossings = OpenKnot(points, verbose=False).raw_crossings().tolist()
        if num_translations == 0:
            return n.array(self_crossings)
        
        # # 2) get crossings of other with other
        # other_crossings = OpenKnot(translated_points).raw_crossings().tolist()
        # if len(other_crossings):
        #     other_crossings[:, :2] += len(points) 

        # 3) get crossings of self with other
        inter_crossings = []
        segment_lengths = (n.roll(translated_points[:, :2], -1, axis=0) -
                           translated_points[:, :2])
        segment_lengths = n.sqrt(n.sum(segment_lengths * segment_lengths,
                                       axis=1))
        max_segment_length = n.max(segment_lengths[:-1])
        jump_mode = 2
        for i in range(len(points) - 1):
            v0 = points[i]
            dv = points[(i+1) % len(points)] - v0

            s = translated_points
            if num_translations > 0:
                vnum = i
                compnum = len(points)
            else:
                vnum = len(points) + i
                compnum = 0

            inter_crossings.extend(chelpers.find_crossings(
                v0, dv, s, segment_lengths,
                vnum, compnum,
                max_segment_length,
                jump_mode
                ))
        # if len(inter_crossings):
        #     inter_crossings = n.array(inter_crossings)
        #     inter_crossings[::2, 1] += len(points)
        #     inter_crossings[1::2, 0] += len(points)
        #     inter_crossings = inter_crossings.tolist()

        if num_translations < 0:
            self_crossings = n.array(self_crossings)
            self_crossings[:, :2] += len(points)
            self_crossings = self_crossings.tolist()

        all_crossings = self_crossings + inter_crossings # + other_crossings
        all_crossings.sort(key=lambda j: j[0])

        return n.array(all_crossings)

    def gauss_code(self, num_translations=0, mat=None, shift=0):
        '''Returns the Gauss code alongside a list of  that are
        crossings of the non-translated curve with the translated one.'''
        crossings = self.raw_crossings(num_translations=num_translations, mat=mat, shift=shift)
        from pyknotid.representations import GaussCode
        gc = GaussCode(crossings)
        if num_translations == 0:
            return gc, gc.crossing_numbers

        admissible_indices = set()
        for line_index, gauss_index in zip(crossings[:, 0], gc._gauss_code[0][:, 0]):
            if num_translations > 0:
                if line_index > len(self.unfolded_points()):
                    admissible_indices.add(gauss_index)
            elif num_translations < 0:
                if line_index < len(self.unfolded_points()):
                    admissible_indices.add(gauss_index)
        return gc, list(admissible_indices)
        

    def plot_projection(self, num_translations=0, mat=None, shift=0):
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        points = self.unfolded_points()
        points = n.vstack((points[shift:], points[:(shift+1)] + self.period_vector()))
        translated_points = points + num_translations * self.period_vector()
        if mat is not None:
            points = n.apply_along_axis(mat.dot, 1, points)
            translated_points = n.apply_along_axis(mat.dot, 1, translated_points)
        ax.plot(points[:, 0], points[:, 1])
        ax.plot(translated_points[:, 0], translated_points[:, 1])
        fig.show()
        return fig, ax

    def plot(self, num_translations=0, mat=None, shift=0):
        import matplotlib.pyplot as plt
        points = self.unfolded_points()
        points = n.vstack((points[shift:], points[:(shift+1)] + self.period_vector()))
        translated_points = points + num_translations * self.period_vector()
        if mat is not None:
            points = n.apply_along_axis(mat.dot, 1, points)
            translated_points = n.apply_along_axis(mat.dot, 1, translated_points)
        from pyknotid.spacecurves import Link
        Link([points, translated_points]).plot()
        
    def periodic_vassiliev_degree_2(self, num_translations=3,
                                    number_of_samples=100,
                                    shift=0):
        from pyknotid.spacecurves.rotation import get_rotation_angles, rotate_to_top            
        angles = get_rotation_angles(number_of_samples)
        v2s = []
        v2_arrs = []
        for translations in range(-1*num_translations, num_translations + 1):
            vprint('Checking translation {}'.format(translations), False)
            v2s_cur = []
            for theta, phi in angles:
                matrix = rotate_to_top(theta + 0.13, phi + 0.43)
                gc, numbers = self.gauss_code(num_translations=translations,
                                              mat=matrix, shift=shift)
                v2s_cur.append(periodic_vassiliev_degree_2(gc, numbers))
                # print(theta + 0.13, phi + 0.43, gc, numbers)
            print()
            print('translation {:03}, v2s_cur {}'.format(translations, v2s_cur))
            v2s.append(n.average(v2s_cur))
            v2_arrs.append(v2s_cur)
        vprint()
        print('totals are', n.sum(v2_arrs, axis=0))
        return n.sum(v2s)

    def periodic_vassiliev_degree_2_by_unfolding(self, num_translations=3, shift=0):
        gc, equivalencies, translation_indices = self.gauss_code_by_unfolding(num_translations, shift=shift)
        from pyknotid.invariants import vassiliev_degree_2
        return periodic_vassiliev_degree_2_without_double_count(gc, equivalencies, translation_indices)
        
    def periodic_vassiliev_degree_3_by_unfolding(self, num_translations=3, shift=0):
        gc = self.gauss_code_by_unfolding(num_translations, shift=shift)[0]
        from pyknotid.invariants import vassiliev_degree_3
        return vassiliev_degree_3(gc)

def periodic_vassiliev_degree_2_without_double_count(representation, equivalent_crossing_numbers={},
                                                     translation_indices={}):
    # Hacky periodic version of the vassiliev function in
    # pyknotid.invariants
    from pyknotid.invariants import _crossing_arrows_and_signs

    # print('gc is', representation)

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate alexander polynomial'
                        'for something with more than 1 component')

    gc = gc[0]
    arrows, signs = _crossing_arrows_and_signs(
        gc, representation.crossing_numbers)

    crossings_already_done = set()
    
    crossing_numbers = list(representation.crossing_numbers)
    representations_sum = 0
    for index, i1 in enumerate(crossing_numbers):
        arrow1 = arrows[i1]
        a1s, a1e = arrow1
        for i2 in crossing_numbers[index+1:]:
            # if (i1 not in relevant_crossing_numbers and
            #     i2 not in relevant_crossing_numbers):
            #     continue

            if tuple(sorted([i1, i2])) in crossings_already_done:
                continue

           # t1 = translation_indices[i1]
            # t2 = translation_indices[i2]
            # non_zero_translations = [t for t in [t1, t2] if t[0] != 0 or t[1] != 0]
            # if len(set(non_zero_translations)) > 1:
            #     continue
            
            arrow2 = arrows[i2]
            a2s, a2e = arrow2

            # if a2s > a1s and a2e < a1s and a1e > a2s:
            if a1e > a2e and a1s < a2e and a2s > a1e:
                

                representations_sum += signs[i1] * signs[i2]
                print('t1')
                print('found with {} and {}'.format(i1, i2))

                # print('vass with', i1, i2, signs[i1] * signs[i2], 'indices',
                #       translation_indices[i1], translation_indices[i2])
                for i1other in equivalent_crossing_numbers[i1].union({i1}):
                    for i2other in equivalent_crossing_numbers[i2].union({i2}):
                        crossings_already_done.add(tuple(sorted([i1other, i2other])))
            elif a1s > a2s and a1e < a2s and a2e > a1s:
                representations_sum += signs[i1] * signs[i2]
                print('t2')
                print('found with {} and {}'.format(i1, i2))

                # print('vass with', i1, i2, '...', signs[i1] * signs[i2], 'indices',
                #       translation_indices[i1], translation_indices[i2])
                for i1other in equivalent_crossing_numbers[i1].union({i1}):
                    for i2other in equivalent_crossing_numbers[i2].union({i2}):
                        crossings_already_done.add(tuple(sorted([i1other, i2other])))

    return representations_sum

def alternative_periodic_vassiliev_degree_2_without_double_count(representation, core_crossings=set(),
                                                                 multiplicities={}, true_crossing_numbers={}):
    # Hacky periodic version of the vassiliev function in
    # pyknotid.invariants
    from pyknotid.invariants import _crossing_arrows_and_signs

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate v2 '
                        'for something with more than 1 component')

    gc = gc[0]
    arrows, signs = _crossing_arrows_and_signs(
        gc, representation.crossing_numbers)

    crossings_done = set()

    crossing_numbers = list(representation.crossing_numbers)
    representations_sum = 0
    for index, i1 in enumerate(crossing_numbers):
        arrow1 = arrows[i1]
        a1s, a1e = arrow1
        for i2 in crossing_numbers[index+1:]:

            arrow2 = arrows[i2]
            a2s, a2e = arrow2

            if not (i1 in core_crossings or i2 in core_crossings):
                continue

            if a1s > a2s and a1e < a2s and a2e > a1s:

                real_cs = tuple(sorted((true_crossing_numbers[i1], true_crossing_numbers[i2])) + [abs(i2 - i1)])
                # real_cs = tuple(sorted((true_crossing_numbers[i1], true_crossing_numbers[i2])))
                if real_cs in crossings_done:
                    continue
                crossings_done.add(real_cs)
                
                representations_sum += signs[i1] * signs[i2]  # * multiplicities[i1] * multiplicities[i2]

    return representations_sum

def alternative_periodic_vassiliev_degree_3_without_double_count(
        representation, core_crossings, true_crossing_numbers):
    # Hacky periodic version of the vassiliev function in
    # pyknotid.invariants
    from pyknotid.invariants import _crossing_arrows_and_signs

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate v2 '
                        'for something with more than 1 component')

    gc = gc[0]
    arrows, signs = _crossing_arrows_and_signs(
        gc, representation.crossing_numbers)

    crossings_done = set()

    crossing_numbers = list(representation.crossing_numbers)
    representations_sum_1 = 0
    representations_sum_2 = 0
    for index, i1 in enumerate(crossing_numbers):
        arrow1 = arrows[i1]
        a1s, a1e = arrow1
        for i2 in crossing_numbers:

            arrow2 = arrows[i2]
            a2s, a2e = arrow2

            for i3 in crossing_numbers:

                arrow3 = arrows[i3]
                a3s, a3e = arrow3

                if not (i1 in core_crossings or i2 in core_crossings or i3 in core_crossings):
                    continue

                # # Not sure if this condition is good
                # if len(set([true_crossing_numbers[ii] for ii in [i1, i2, i3]])) < 3:
                #     continue

                if len(set([i1, i2, i3])) < 3:
                    continue

                # if len(set([true_crossing_numbers[ii] for ii in [i1, i2, i3]])) < 3:
                #     continue

                si1, si2, si3 = sorted([i1, i2, i3])
                real_cs = tuple(
                    sorted(
                        (true_crossing_numbers[i1],
                         true_crossing_numbers[i2],
                         true_crossing_numbers[i3])) + [
                             si2 - si1,
                             si3 - si2] + [(a1s < a2s < a3e < a1e < a3s < a2e),
                                           (a1s < a2e < a3e < a2s < a1e < a3s),
                                           (a1e < a2e < a1s < a3e < a2s < a3s),
                                           (a1e < a2s < a3e < a1s < a3s < a2e),
                                           (a1s < a2e < a3s < a2s < a1e < a3e),
                                           (a1e < a2s < a1s < a3e < a2e < a3s)])

                if i1 == 20 and i2 == 21 and i3 == 22:
                    print('!!!')
                    print('rcs', real_cs)
                    print((a1s < a2s < a3e < a1e < a3s < a2e),
                          (a1s < a2e < a3e < a2s < a1e < a3s),
                          (a1e < a2e < a1s < a3e < a2s < a3s),
                          (a1e < a2s < a3e < a1s < a3s < a2e),
                          (a1s < a2e < a3s < a2s < a1e < a3e),
                          (a1e < a2s < a1s < a3e < a2e < a3s))
                    

                if real_cs in crossings_done:
                    continue

                if ((a1s < a2s < a3e < a1e < a3s < a2e) or
                    (a1s < a2e < a3e < a2s < a1e < a3s) or
                    (a1e < a2e < a1s < a3e < a2s < a3s) or
                    (a1e < a2s < a3e < a1s < a3s < a2e) or
                    (a1s < a2e < a3s < a2s < a1e < a3e) or
                    (a1e < a2s < a1s < a3e < a2e < a3s)):
                    print('r1 with', i1, i2, i3, signs[i1] * signs[i2] *
                          signs[i3])
                    print('and real cs', real_cs)
                    print((a1s < a2s < a3e < a1e < a3s < a2e),
                          (a1s < a2e < a3e < a2s < a1e < a3s),
                          (a1e < a2e < a1s < a3e < a2s < a3s),
                          (a1e < a2s < a3e < a1s < a3s < a2e),
                          (a1s < a2e < a3s < a2s < a1e < a3e),
                          (a1e < a2s < a1s < a3e < a2e < a3s))
                    representations_sum_1 += (signs[i1] * signs[i2] *
                                              signs[i3])
                    crossings_done.add(real_cs)
                # if ((a1s < a2e < a3s < a1e < a3e < a2s) or
                #     (a1e < a2s < a3e < a2e < a1s < a3s) or
                #     (a1s < a2e < a1e < a3s < a2s < a3e) or
                #     (a1e < a2e < a3s < a1s < a3e < a2s) or
                #     (a1e < a2s < a3s < a2e < a1s < a3e) or
                #     (a1s < a2s < a1e < a3s < a2e < a3e)):
                #     print('r1 alternative with', i1, i2, i3, signs[i1] * signs[i2] *
                #           signs[i3])
                #     print('and real cs', real_cs)
                #     print((a1s < a2s < a3e < a1e < a3s < a2e),
                #           (a1s < a2e < a3e < a2s < a1e < a3s),
                #           (a1e < a2e < a1s < a3e < a2s < a3s),
                #           (a1e < a2s < a3e < a1s < a3s < a2e),
                #           (a1s < a2e < a3s < a2s < a1e < a3e),
                #           (a1e < a2s < a1s < a3e < a2e < a3s))
                #     representations_sum_1 += (signs[i1] * signs[i2] *
                #                               signs[i3])
                #     crossings_done.add(real_cs)
                if ((a2s > a1e and a3e > a2s and a1s > a3e and a2e > a1s and a3s > a2e) or
                    (a2e > a1s and a3s > a2e and a1e > a3s and a2s > a1e and a3e > a2s)):
                  
                    print('r3 with', i1, i2, i3, signs[i1] * signs[i2] *
                          signs[i3])
                    print('and real cs', real_cs)
                    representations_sum_2 += (signs[i1] * signs[i2] *
                                              signs[i3])
                    crossings_done.add(real_cs)

    return representations_sum_1 / 2. + representations_sum_2
    return int(round(representations_sum_1 / 2.)) + representations_sum_2


def periodic_vassiliev_degree_3_without_double_count(representation,
                                                     equivalent_crossing_numbers,
                                                     translations):
    ## See Polyak and Viro
    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate alexander polynomial'
                        'for something with more than 1 component')

    gc = gc[0]
    from pyknotid.invariants import _crossing_arrows_and_signs
    arrows, signs = _crossing_arrows_and_signs(
        gc, representation.crossing_numbers)
    
    crossings_already_done = set()

    crossing_numbers = list(representation.crossing_numbers)
    used_sets = set()
    representations_sum_1 = 0
    representations_sum_2 = 0
    for index, i1 in enumerate(crossing_numbers):
        # if index % 10 == 0:
        #     vprint('\rCurrently comparing index {}'.format(index), False)
        arrow1 = arrows[i1]
        a1s, a1e = arrow1
        a1e = (a1e - a1s) % len(gc)
        for i2 in crossing_numbers:
            arrow2 = arrows[i2]
            a2s, a2e = arrow2
            a2s = (a2s - a1s) % len(gc)
            a2e = (a2e - a1s) % len(gc)
            for i3 in crossing_numbers:
                if tuple(sorted([i1, i2, i3])) in crossings_already_done:
                    continue
                
                arrow3 = arrows[i3]
                a3s, a3e = arrow3
                a3s = (a3s - a1s) % len(gc)
                a3e = (a3e - a1s) % len(gc)

                # t1 = translations[i1]
                # t2 = translations[i2]
                # t3 = translations[i3]
                # non_zero_translations = [t for t in [t1, t2, t3] if t[0] != 0 or t[1] != 0]
                # if len(set(non_zero_translations)) > 1:
                #     # print('translations invalid', t1, t2, t3)
                #     continue

                ordered_indices = tuple(sorted((i1, i2, i3)))
                if ordered_indices in used_sets:
                    continue

                if (a2s < a1e and a3e < a1e and a3e > a2s and
                    a3s > a1e and a2e > a3s):
                    representations_sum_1 += (signs[i1] * signs[i2] *
                                              signs[i3])
                    used_sets.add(ordered_indices)

                    # print('vass 1 with', i1, i2, i3, '...', signs[i1] * signs[i2] * signs[i3], 'indices',
                    #       translations[i1], translations[i2], translations[i3])

                    for i1other in equivalent_crossing_numbers[i1].union({i1}):
                        for i2other in equivalent_crossing_numbers[i2].union({i2}):
                            for i3other in equivalent_crossing_numbers[i3].union({i3}):
                                crossings_already_done.add(tuple(sorted([i1other, i2other, i3other])))

                if (a2e < a1e and a3s < a1e and a3s > a2e and
                    a2s > a1e and a3e > a2s):
                    representations_sum_2 += (signs[i1] * signs[i2] *
                                              signs[i3])
                    used_sets.add(ordered_indices)

                    # print('vass 2 with', i1, i2, i3, '...', signs[i1] * signs[i2] * signs[i3], 'indices',
                    #       translations[i1], translations[i2], translations[i3])

                    for i1other in equivalent_crossing_numbers[i1].union({i1}):
                        for i2other in equivalent_crossing_numbers[i2].union({i2}):
                            for i3other in equivalent_crossing_numbers[i3].union({i3}):
                                crossings_already_done.add(tuple(sorted([i1other, i2other, i3other])))

    # print()
    
    return representations_sum_1 / 2. + representations_sum_2

def periodic_vassiliev_degree_2(representation, relevant_crossing_numbers=[]):
    # Hacky periodic version of the vassiliev function in
    # pyknotid.invariants
    from pyknotid.invariants import _crossing_arrows_and_signs

    relevant_crossing_numbers = set(relevant_crossing_numbers)

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate alexander polynomial'
                        'for something with more than 1 component')

    gc = gc[0]
    arrows, signs = _crossing_arrows_and_signs(
        gc, representation.crossing_numbers)
    
    crossing_numbers = list(representation.crossing_numbers)
    representations_sum = 0
    for index, i1 in enumerate(crossing_numbers):
        arrow1 = arrows[i1]
        a1s, a1e = arrow1
        for i2 in crossing_numbers[index+1:]:
            if (i1 not in relevant_crossing_numbers and
                i2 not in relevant_crossing_numbers):
                continue
            arrow2 = arrows[i2]
            a2s, a2e = arrow2

            if a2s > a1s and a2e < a1s and a1e > a2s:
                representations_sum += signs[i1] * signs[i2]
            elif a1s > a2s and a1e < a2s and a2e > a1s:
                representations_sum += signs[i1] * signs[i2]

    return representations_sum

def _fold(points, shape):
    '''Imposes the shape as periodic boundary conditions.'''
    shape = n.array(shape)
    dx, dy, dz = shape

    points = n.vstack([[points[0] + 0.001], points])
    rest = n.array(points).copy()
    new_points = []
    i = 0
    while i < len(rest) - 1:
        cur = rest[i]
        nex = rest[i+1]
        print('i is', i)
        print(cur, nex)
        if nex[0] < 0:
            new_points.append(rest[:i+1])
            boundary_point = cur + cur[0] / (cur[0] - nex[0]) * (nex - cur)
            new_points.append([boundary_point])
            rest = n.vstack([[boundary_point], rest[i+1:]])
            rest[:, 0] += dx
            i = 0
        elif nex[0] > dx:
            new_points.append(rest[:i+1])
            boundary_point = cur + (dx - cur[0]) / (nex[0] - cur[0]) * (nex - cur)
            new_points.append([boundary_point])
            rest = n.vstack([[boundary_point], rest[i+1:]])
            rest[:, 0] -= dx
            i = 0
        elif nex[1] < 0:
            new_points.append(rest[:i+1])
            boundary_point = cur + cur[1] / (cur[1] - nex[1]) * (nex - cur)
            new_points.append([boundary_point])
            rest = n.vstack([[boundary_point], rest[i+1:]])
            rest[:, 1] += dy
            i = 0
        elif nex[1] > dy:
            new_points.append(rest[:i+1])
            boundary_point = cur + (dx - cur[1]) / (nex[1] - cur[1]) * (nex - cur)
            new_points.append([boundary_point])
            rest = n.vstack([[boundary_point], rest[i+1:]])
            rest[:, 1] -= dy
            i = 0
        elif nex[2] < 0:
            new_points.append(rest[:i+1])
            boundary_point = cur + cur[2] / (cur[2] - nex[2]) * (nex - cur)
            new_points.append([boundary_point])
            rest = n.vstack([[boundary_point], rest[i+1:]])
            rest[:, 2] += dz
            i = 0
        elif nex[2] > dz:
            new_points.append(rest[:i+1])
            boundary_point = cur + (dx - cur[2]) / (nex[2] - cur[2]) * (nex - cur)
            new_points.append([boundary_point])
            rest = n.vstack([[boundary_point], rest[i+1:]])
            rest[:, 2] -= dz
            i = 0
        else:
            i += 1
    new_points.append(rest)

    print('new points', new_points)

    return n.vstack(new_points)[1:]
    
def get_true_crossing_numbers(equivalent_cs, cores):
    output = {}
    for crossing, equivalents in equivalent_cs.items():
        output[crossing] = min([c for c in equivalents if c in cores])

    return output


def alternative_periodic_vassiliev_degree_4_conway_z4(
        representation, core_crossings, true_crossing_numbers):
    # Hacky periodic version of the vassiliev function in
    # pyknotid.invariants
    from pyknotid.invariants import _crossing_arrows_and_signs

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate v2 '
                        'for something with more than 1 component')

    gc = gc[0]
    arrows, signs = _crossing_arrows_and_signs(
        gc, representation.crossing_numbers)

    crossings_done = set()

    crossing_numbers = list(representation.crossing_numbers)
    representations_sum = 0
    for index, i1 in enumerate(crossing_numbers):
        arrow1 = arrows[i1]
        a1s, a1e = arrow1
        for i2 in crossing_numbers:

            arrow2 = arrows[i2]
            a2s, a2e = arrow2

            for i3 in crossing_numbers:

                arrow3 = arrows[i3]
                a3s, a3e = arrow3

                for i4 in crossing_numbers:
                    arrow4 = arrows[i4]
                    a4s, a4e = arrow4

                    if not (i1 in core_crossings or i2 in core_crossings or i3 in core_crossings or i4 in core_crossings):
                        continue

                    if len(set([true_crossing_numbers[ii] for ii in (i1, i2, i3, i4)])) < 4:
                        continue
                    si1, si2, si3, si4 = sorted([i1, i2, i3, i4])
                    real_cs = tuple(
                        sorted(
                            (true_crossing_numbers[i1],
                             true_crossing_numbers[i2],
                             true_crossing_numbers[i3],
                             true_crossing_numbers[i4])) + [
                                 si2 - si1,
                                 si3 - si2,
                                 si4 - si3] + [(a1s < a2e < a3s < a4e < a1e < a2s < a3e < a4s),
                                               (a1s < a2e < a1e < a2s < a3s < a4e < a3e < a4s),
                                               (a1s < a2e < a1e < a3s < a4e < a3e < a4s < a1s),
                                               (a1s < a2e < a3s < a4e < a3e < a4s < a1e < a2s),
                                               (a1s < a2s < a3e < a2e < a3s < a4e < a1e < a4s),
                                               #
                                               (a1s < a2e < a3s < a4e < a3e < a1e < a2s < a4s),
                                               (a1s < a2e < a3e < a2s < a4e < a1e < a3s < a4s),
                                               (a1s < a2e < a1e < a3s < a4e < a2s < a3e < a4s),
                                               (a1s < a2e < a3e < a4s < a1e < a3s < a4e < a2s),
                                               (a1s < a2e < a3s < a4e < a2s < a3e < a1e < a4s),
                                               (a1s < a2e < a4s < a1e < a3s < a4e < a3e < a4s),
                                               (a1s < a2s < a3e < a1e < a4s < a2e < a4e < a3s),
                                               (a1s < a2s < a3e < a4e < a1e < a4s < a2e < a3s),
                                               #
                                               (a1s < a2e < a1e < a3s < a2s < a4e < a3e < a4s),
                                               (a1s < a2e < a3s < a1e < a4s < a3e < a4e < a2s),
                                               (a1s < a2s < a3e < a4s < a2e < a4e < a1e < a3e),
                                               (a1s < a2s < a3e < a1e < a3s < a4e < a2e < a4s),
                                               (a1s < a2s < a3e < a2e < a4e < a1e < a4s < a3s),
                                               (a1s < a2e < a1e < a3s < a4e < a3e < a2s < a4s),
                                               (a1s < a2e < a3e < a4e < a3s < a1e < a4s < a2s),
                                               (a1s < a2e < a3e < a2s < a4e < a3s < a1e < a4s)])

                    if real_cs in crossings_done:
                        continue

                    # if ((a2e > a1s and a3s > a2e and a4e > a3s and a1e > a4e and a3s > a1e and a3e > a2s and a4s > a3e) or
                    #     (a2e > a1s and a1e > a2e and a2s > a1e and a3s > a2s and a4e > a3s and a3e > a4e and a4s > a3e) or
                    #     (a2e > a1s and a1e > a2e and a3s > a1e and a4e > a3s and a3e > a4e and a4s > a3e and a2s > a4s) or
                    #     (a2e > a1s and a3s > a2e and a4e > a3s and a3e > a4e and a4s > a3e and a1e > a4s and a2s > a1e) or
                    #     (a2s > a1s and a3e > a2s and a2e > a3e and a3s > a2e and a4e > a3s and a1e > a4e and a4s > a1e) or
                    if ((a1s < a2e < a3s < a4e < a1e < a2s < a3e < a4s) or
                        (a1s < a2e < a1e < a2s < a3s < a4e < a3e < a4s) or
                        (a1s < a2e < a1e < a3s < a4e < a3e < a4s < a1s) or
                        (a1s < a2e < a3s < a4e < a3e < a4s < a1e < a2s) or
                        (a1s < a2s < a3e < a2e < a3s < a4e < a1e < a4s) or
                        #
                        (a1s < a2e < a3s < a4e < a3e < a1e < a2s < a4s) or
                        (a1s < a2e < a3e < a2s < a4e < a1e < a3s < a4s) or
                        (a1s < a2e < a1e < a3s < a4e < a2s < a3e < a4s) or
                        (a1s < a2e < a3e < a4s < a1e < a3s < a4e < a2s) or
                        (a1s < a2e < a3s < a4e < a2s < a3e < a1e < a4s) or
                        (a1s < a2e < a4s < a1e < a3s < a4e < a3e < a4s) or
                        (a1s < a2s < a3e < a1e < a4s < a2e < a4e < a3s) or
                        (a1s < a2s < a3e < a4e < a1e < a4s < a2e < a3s) or
                        #
                        (a1s < a2e < a1e < a3s < a2s < a4e < a3e < a4s) or
                        (a1s < a2e < a3s < a1e < a4s < a3e < a4e < a2s) or
                        (a1s < a2s < a3e < a4s < a2e < a4e < a1e < a3e) or
                        (a1s < a2s < a3e < a1e < a3s < a4e < a2e < a4s) or
                        (a1s < a2s < a3e < a2e < a4e < a1e < a4s < a3s) or
                        (a1s < a2e < a1e < a3s < a4e < a3e < a2s < a4s) or
                        (a1s < a2e < a3e < a4e < a3s < a1e < a4s < a2s) or
                        (a1s < a2e < a3e < a2s < a4e < a3s < a1e < a4s)):
                        
                            print('r with', i1, i2, i3, i4, signs[i1] * signs[i2] *
                                  signs[i3] * signs[i4])
                            # print('true', [true_crossing_numbers[ii] for ii in (i1, i2, i3, i4)])
                            representations_sum += (signs[i1] * signs[i2] *
                                                    signs[i3] * signs[i4])
                            crossings_done.add(real_cs)

    return representations_sum



'''
Cython functions for space curve analysis.
'''

import numpy as n
cimport numpy as n

cimport cython

from libc.math cimport abs, pow, sqrt as csqrt, floor

cpdef find_crossings(double [:] v, double [:] dv,
                     double [:, :] points,
                     double [:] segment_lengths,
                     long current_index,
                     long comparison_index,
                     double max_segment_length,
                     long jump_mode=1
                     ):
    '''
    Searches for crossings between the given vector and any other
    vector in the
    list of points, returning all of them as a list.
                     
    Parameters
    ----------
    v0 : ndarray
        The current point, a 1D vector.
    dv : ndarray
        The vector connecting the current point to the next one
    points : ndarray
        The array or (x, y) values of all the other points
    segment_lengths : ndarray
        The length of each segment joining a point to the
        next one.
    current_index : long
        The index of the point currently being tested.
    comparison_index : long
        The index of the first comparison point
    jump_mode : int
        1 to check every jump distance, 2 to jump based on
        the maximum one, 3 to never jump and check the length
        of every step.
    '''

    cdef list crossings = []
    cdef double vx = v[0]
    cdef double vy = v[1]
    cdef double vz = v[2]
    cdef double dvx = dv[0]
    cdef double dvy = dv[1]
    cdef double dvz = dv[2]

    cdef double twice_max_segment_length = 2*max_segment_length

    cdef int i = 0
    cdef double distance, distance_travelled
    cdef double [:] point
    cdef double [:] next_point
    cdef double jump_x, jump_y, jump_z
    cdef double pz
    cdef double dpz
    cdef long intersect
    cdef double intersect_i, intersect_j
    cdef double crossing_sign
    cdef double crossing_direction
    cdef long jumps
    cdef long num_jumps
    cdef int already_jumped = 0

    while i < len(points) - 1:
        point = points[i]
        distance = csqrt(pow(vx - point[0], 2) + pow(vy - point[1], 2))
        if distance < twice_max_segment_length or already_jumped:
            already_jumped = 0
            next_point = points[i+1]
            jump_x = next_point[0] - point[0]
            jump_y = next_point[1] - point[1]
            jump_z = next_point[2] - point[2]


            intersect, intersect_i, intersect_j = do_vectors_intersect(
                vx, vy, dvx, dvy, point[0], point[1],
                jump_x, jump_y)

            if intersect:
                pz = point[2]
                dpz = jump_z

                crossing_sign = sign((vz + intersect_i * dvz) -
                                     (pz + intersect_j * dpz))

                crossing_direction = sign(cross_product(
                    dvx, dvy, jump_x, jump_y))

                crossings.append([<double>current_index + intersect_i,
                                  (<double>comparison_index + intersect_j +
                                   <double>i),
                                  crossing_sign,
                                  crossing_sign * crossing_direction])
                crossings.append([(<double>comparison_index + intersect_j +
                                   <double>i),
                                  <double>current_index + intersect_i,
                                  -1. * crossing_sign,
                                  crossing_sign * crossing_direction])
            i += 1

        elif jump_mode == 3:
            i += 1  # naive mode - check everything
            already_jumped = 1
        elif jump_mode == 2:
            num_jumps = <long>(floor(distance / max_segment_length)) - 1
            if num_jumps < 1:
                num_jumps = 1
            i += num_jumps
            already_jumped = 1
        else:  # Catch all other jump modes
            distance_travelled = 0.
            jumps = 0
            while (distance_travelled < (distance - max_segment_length) and
                   i < len(points)):
                jumps += 1
                distance_travelled += segment_lengths[i]
                i += 1
            if jumps > 1:
                i -= 2
            already_jumped = 1
            # This keeps jumping until we might be close enough to intersect,
            # without doing vector arithmetic at every step
                                  
    return crossings
                                      



cdef tuple do_vectors_intersect(double px, double py, double dpx, double dpy,
                                double qx, double qy, double dqx, double dqy):
    """Takes four vectors p, dp and q, dq, then tests whether they cross in
    the dp/dq region. Returns this boolean, and the (fractional) point where
    the crossing actually occurs.
    """
    cdef double t, u

    if abs(cross_product(dpx, dpy, dqx, dqy)) < 0.000001:
        return (0, 0., 0.)

    t = cross_product(qx - px, qy - py, dqx, dqy) / cross_product(dpx, dpy,
                                                                  dqx, dqy)
    if t < 1.0 and t > 0.0:
        u = cross_product(qx - px, qy - py, dpx, dpy) / cross_product(dpx, dpy,
                                                                      dqx, dqy)
        if u < 1.0 and u > 0.0:
            return (1, t, u)
    return (0, -1., -1.)
        
cpdef double cross_product(double px, double py, double qx, double qy):
    '''Simple cython cross product for 2D vectors.'''
    return px * qy - py * qx

cpdef double sign(double a):
    return (1. if a > 0. else (-1. if a < 0. else 0.))

cpdef double mag_difference(double [:] a, double [:] b):
    '''The magnitude of the vector joining a and b'''
    return csqrt((b[0] - a[0])**2 + (b[1] - a[1])**2)



'''
Knots and links can be encoded in many different ways, generally by
enumerating their self-intersections in projection along some
axis. We provide here

This module contains classes and functions for representing
knots in knot diagrams, mainly the
:class:`pyknotid.representations.gausscode.GaussCode` and
:class:`pyknotid.representations.planardiagram.PlanarDiagram`.

These provide convenient methods to convert between different
representations, and to simplify via Reidemeister moves.

Creating representations
########################

Knot representations can be calculated from space curves, or created directly by inputting standard notations.

From space curves
-----------------

pyknotid's space curve classes can all return topological representations. For instance::

  from pyknotid.spacecurves import Knot
  from pyknotid.make import trefoil
  k = Knot(trefoil())
  
You can extract a :class:`~pyknotid.representations.gausscode.GaussCode` object::

  k.gauss_code()  # 1+a,2-a,3+a,1-a,2+a,3-a

or a :class:`~pyknotid.representations.planardiagram.PlanarDiagram`::

  k.planar_diagram()  # PD with 3: X_{2,5,3,6} X_{4,1,5,2} X_{6,3,1,4}

or a Gauss diagram::

  k.gauss_diagram()  # plots the diagram in a new window using matplotlib

.. image:: example_gauss_diagram_k10_93.png
   :alt: An example Gauss diagram for the knot 10_93
   :align: center
   :scale: 50%

or a generic :class:`~pyknotid.representations.representation.Representation`::

  k.representation()  # 1+a,2-a,3+a,1-a,2+a,3-a (but provides more methods than a GaussCode)

By direct input
---------------

Gauss codes
~~~~~~~~~~~

A Gauss code is a list of crossings in a projection of a curve,
labelled by numbers, and in each case indicating whether the curve
passes over (+) or under (-) itself. Each crossing also has a local
orientation, represented here by 'c' for clockwise, or 'a' for
anticlockwise.

With these rules, you can enter Gauss codes as comma-separated lists::

  from pyknotid.representations import GaussCode
  gc = GaussCode('1+c,2-c,3+c,1-c,2+c,3-c')

If you do not know the crossing orientations (c/a), pyknotid can
calculate them automatically::

  gc = GaussCode.calculating_orientations('1+,2-,3+,1-,2+,3-')

If you do this with a chiral not, the chirality is selected
arbitrarily.


Calculating invariants
######################

You can calculate many invariants using the functions of
:doc:`../invariants`.

pyknotid also provides a more convenient interface using the
:class:`~pyknotid.representations.representation.Representation`
class. Internally this wraps a Gauss code::

  from pyknotid.representations import Representation
  rep = Representation('1-c,2+c,3-a,4+a,2-c,1+c,4-a,3+a')

You can then calculate many quantities via methods of this object::

  rep.vassiliev_degree_2()  # 1
  rep.vassiliev_degree_3()  # -1
  rep.identify()  # [<Knot 4_1>]

For a full list of available functions, see
:class:`~pyknotid.representations.representation.Representation`.

'''

from pyknotid.representations.gausscode import GaussCode
from pyknotid.representations.planardiagram import PlanarDiagram
from pyknotid.representations.dtnotation import DTNotation
from pyknotid.representations.representation import Representation
__all__ = ['GaussCode', 'PlanarDiagram', 'DTNotation', 'Representation', ]

